from edsnlp.utils.regex import make_pattern

covid = [
    r"covid([-\s]?19)?",
    r"sars[-\s]?cov[-\s]?2",
    r"corona[-\s]?virus",
]

diseases = [r"pneumopathies?", r"infections?"]

patterns = [r"(" + make_pattern(diseases) + r"\s[àa]u?\s)?" + make_pattern(covid)]



main_pattern = dict(
    source="main",
    regex=[
        r"ulcere.{1,10}gastr",
        r"ulcere.{1,10}duoden",
        r"ulcere.{1,10}antra",
        r"ulcere.{1,10}pept",
        r"ulcere.{1,10}estomac",
        r"ulcere.{1,10}curling",
        r"ulcere.{1,10}bulb",
        r"(œ|oe)sophagites.{1,5}pepti.{1,10}ulcer",
        r"gastrite.{1,20}ulcer",
        r"antrite.{1,5}ulcer",
    ],
    regex_attr="NORM",
)

acronym = dict(
    source="acronym",
    regex=[
        r"\bUGD\b",
    ],
    regex_attr="TEXT",
)

generic = dict(
    source="generic",
    regex=r"ulcere",
    regex_attr="NORM",
    assign=dict(
        name="is_peptic",
        regex=r"\b(gastr|digest)",
        window=(-20, 20),
        limit_to_sentence=False,
    ),
)

default_patterns = [
    main_pattern,
    acronym,
    generic,
]



number_terms = {
    "0.125": ["⅛"],
    "0.16666666": ["⅙"],
    "0.2": ["⅕"],
    "0.25": ["¼"],
    "0.3333333": ["⅓"],
    "0.5": ["½"],
    "2.5": ["21/2"],
    "1": ["un", "une"],
    "2": ["deux"],
    "3": ["trois"],
    "4": ["quatre"],
    "5": ["cinq"],
    "6": ["six"],
    "7": ["sept"],
    "8": ["huit"],
    "9": ["neuf"],
    "10": ["dix"],
    "11": ["onze"],
    "12": ["douze"],
    "13": ["treize"],
    "14": ["quatorze"],
    "15": ["quinze"],
    "16": ["seize"],
    "17": ["dix-sept", "dix sept"],
    "18": ["dix-huit", "dix huit"],
    "19": ["dix-neuf", "dix neuf"],
    "20": ["vingt", "vingts"],
    "30": ["trente"],
    "40": ["quarante"],
    "50": ["cinquante"],
    "60": ["soixante"],
    "70": ["soixante dix", "soixante-dix"],
    "80": ["quatre vingt", "quatre-vingt", "quatre vingts", "quatre-vingts"],
    "90": ["quatre vingt dix", "quatre-vingt-dix"],
    "100": ["cent"],
    "500": ["cinq cent", "cinq-cent"],
    "1000": ["mille", "milles"],
}


number_regex = r"""(?x)
# no digit or floating point number prefix before
(?<![0-9][.,]?)
# integer part like 123 or 1 234
(?:
    0
    |[1-9][0-9]*(?:\ \d{3})*
)
(?:
    # floating point surounded by spaces
    \ +[,.]\ +\d+
    # floating point w/o space
    | [,.]\d+
    # fractions
    | (?:\ /\ |/)[1-9][0-9]*(?:\ \d{3})*
)?"""


units_config = {
    # Lengths
    "µm": {
        "dim": "length",
        "degree": 1,
        "scale": 1e-4,
        "terms": [
            "micrometre",
            "micrometres",
            "micro-metre",
            "micrometres",
            "µm",
            "um",
        ],
        "followed_by": None,
    },
    "mm": {
        "dim": "length",
        "degree": 1,
        "scale": 1e-1,
        "terms": ["millimetre", "millimetres", "milimetre", "milimetres", "mm"],
        "followed_by": None,
    },
    "cm": {
        "dim": "length",
        "degree": 1,
        "scale": 1e0,
        "terms": ["centimetre", "centimetres", "cm"],
        "followed_by": None,
    },
    "dm": {
        "dim": "length",
        "degree": 1,
        "scale": 1e1,
        "terms": ["decimetre", "decimetres", "dm"],
        "followed_by": None,
    },
    "m": {
        "dim": "length",
        "degree": 1,
        "scale": 1e2,
        "terms": ["metre", "metres", "m"],
        "followed_by": "cm",
    },
    # Weights
    "µg": {
        "dim": "mass",
        "degree": 1,
        "scale": 1e-3,
        "terms": [
            "microgramme",
            "microgrammes",
            "micro-gramme",
            "microgrammes",
            "µg",
        ],
        "followed_by": None,
    },
    "mg": {
        "dim": "mass",
        "degree": 1,
        "scale": 1e0,
        "terms": [
            "milligramme",
            "miligramme",
            "milligrammes",
            "miligrammes",
            "mgr",
            "mg",
        ],
        "followed_by": None,
    },
    "cg": {
        "dim": "mass",
        "degree": 1,
        "scale": 1e1,
        "terms": ["centigramme", "centigrammes", "cg", "cgr"],
        "followed_by": None,
    },
    "dg": {
        "dim": "mass",
        "degree": 1,
        "scale": 1e2,
        "terms": ["decigramme", "decigrammes", "dgr", "dg"],
        "followed_by": None,
    },
    "g": {
        "dim": "mass",
        "degree": 1,
        "scale": 1e3,
        "terms": ["gramme", "grammes", "gr", "g"],
        "followed_by": None,
    },
    "kg": {
        "dim": "mass",
        "degree": 1,
        "scale": 1e6,
        "terms": ["kilo", "kilogramme", "kilogrammes", "kgr", "kg"],
        "followed_by": "g",
    },
    # Durations
    "second": {
        "dim": "time",
        "degree": 1,
        "scale": 1,
        "terms": ["seconde", "secondes", "s"],
        "followed_by": None,
    },
    "minute": {
        "dim": "time",
        "degree": 1,
        "scale": 60,
        "terms": ["mn", "min", "minute", "minutes"],
        "followed_by": "second",
    },
    "hour": {
        "dim": "time",
        "degree": 1,
        "scale": 3600,
        "terms": ["heure", "heures", "h"],
        "followed_by": "minute",
    },
    "day": {
        "dim": "time",
        "degree": 1,
        "scale": 3600 * 24,
        "terms": ["jour", "jours", "j"],
        "followed_by": None,
    },
    "month": {
        "dim": "time",
        "degree": 1,
        "scale": 3600 * 24 * 30.4167,
        "terms": ["mois"],
        "followed_by": None,
    },
    "week": {
        "dim": "time",
        "degree": 1,
        "scale": 3600 * 24 * 7,
        "terms": ["semaine", "semaines", "sem"],
        "followed_by": None,
    },
    "year": {
        "dim": "time",
        "degree": 1,
        "scale": 3600 * 24 * 365.25,
        "terms": ["an", "année", "ans", "années"],
        "followed_by": None,
    },
    # Angle
    "arc-second": {
        "dim": "time",
        "degree": 1,
        "scale": 2 / 60.0,
        "terms": ['"', "''"],
        "followed_by": None,
    },
    "arc-minute": {
        "dim": "time",
        "degree": 1,
        "scale": 2,
        "terms": ["'"],
        "followed_by": "arc-second",
    },
    "degree": {
        "dim": "time",
        "degree": 1,
        "scale": 120,
        "terms": ["degre", "°", "deg"],
        "followed_by": "arc-minute",
    },
    # Temperature
    "celsius": {
        "dim": "temperature",
        "degree": 1,
        "scale": 1,
        "terms": ["°C", "° celsius", "celsius"],
        "followed_by": None,
    },
    # Volumes
    "ml": {
        "dim": "length",
        "degree": 3,
        "scale": 1e0,
        "terms": ["mililitre", "millilitre", "mililitres", "millilitres", "ml"],
        "followed_by": None,
    },
    "cl": {
        "dim": "length",
        "degree": 3,
        "scale": 1e1,
        "terms": ["centilitre", "centilitres", "cl"],
        "followed_by": None,
    },
    "dl": {
        "dim": "length",
        "degree": 3,
        "scale": 1e2,
        "terms": ["decilitre", "decilitres", "dl"],
        "followed_by": None,
    },
    "l": {
        "dim": "length",
        "degree": 3,
        "scale": 1e3,
        "terms": ["litre", "litres", "l", "dm3"],
        "followed_by": "ml",
    },
    "cac": {
        "dim": "length",
        "degree": 3,
        "scale": 5e0,
        "terms": ["cac", "c.a.c", "cuillere à café", "cuillères à café"],
        "followed_by": None,
    },
    "goutte": {
        "dim": "length",
        "degree": 3,
        "scale": 5e-2,
        "terms": ["gt", "goutte", "gouttes"],
        "followed_by": None,
    },
    "mm3": {
        "dim": "length",
        "degree": 3,
        "scale": 1e-3,
        "terms": ["mm3", "mm³"],
        "followed_by": None,
    },
    "cm3": {
        "dim": "length",
        "degree": 3,
        "scale": 1e0,
        "terms": ["cm3", "cm³", "cc"],
        "followed_by": None,
    },
    "dm3": {
        "dim": "length",
        "degree": 3,
        "scale": 1e3,
        "terms": ["dm3", "dm³"],
        "followed_by": None,
    },
    "m3": {
        "dim": "length",
        "degree": 3,
        "scale": 1e6,
        "terms": ["m3", "m³"],
        "followed_by": None,
    },
    # Surfaces
    "µm2": {
        "dim": "length",
        "degree": 2,
        "scale": 1e-8,
        "terms": ["µm2", "µm²"],
        "followed_by": None,
    },
    "mm2": {
        "dim": "length",
        "degree": 2,
        "scale": 1e-2,
        "terms": ["mm2", "mm²"],
        "followed_by": None,
    },
    "cm2": {
        "dim": "length",
        "degree": 2,
        "scale": 1e0,
        "terms": ["cm2", "cm²"],
        "followed_by": None,
    },
    "dm2": {
        "dim": "length",
        "degree": 2,
        "scale": 1e2,
        "terms": ["dm2", "dm²"],
        "followed_by": None,
    },
    "m2": {
        "dim": "length",
        "degree": 2,
        "scale": 1e4,
        "terms": ["m2", "m²"],
        "followed_by": None,
    },
    # International units
    "mui": {
        "dim": "ui",
        "degree": 1,
        "scale": 1e0,
        "terms": ["mui", "m ui"],
        "followed_by": None,
    },
    "dui": {
        "dim": "ui",
        "degree": 1,
        "scale": 1e1,
        "terms": ["dui", "d ui"],
        "followed_by": None,
    },
    "cui": {
        "dim": "ui",
        "degree": 1,
        "scale": 1e2,
        "terms": ["cui", "c ui"],
        "followed_by": None,
    },
    "ui": {
        "dim": "ui",
        "degree": 1,
        "scale": 1e3,
        "terms": ["ui"],
        "followed_by": None,
    },
    # Inverse
    "per_µm": {
        "dim": "length",
        "degree": -1,
        "scale": 1e4,
        "terms": ["µm-1"],
        "followed_by": None,
    },
    "per_mm": {
        "dim": "length",
        "degree": -1,
        "scale": 1e1,
        "terms": ["mm-1"],
        "followed_by": None,
    },
    "per_cm": {
        "dim": "length",
        "degree": -1,
        "scale": 1e0,
        "terms": ["cm-1"],
        "followed_by": None,
    },
    "per_dm": {
        "dim": "length",
        "degree": -1,
        "scale": 1e-1,
        "terms": ["dm-1"],
        "followed_by": None,
    },
    "per_m": {
        "dim": "length",
        "degree": -1,
        "scale": 1e-2,
        "terms": ["m-1"],
        "followed_by": None,
    },
    "per_mg": {
        "dim": "mass",
        "degree": -1,
        "scale": 1e-0,
        "terms": ["mgr-1", "mg-1", "mgr⁻¹", "mg⁻¹"],
        "followed_by": None,
    },
    "per_cg": {
        "dim": "mass",
        "degree": -1,
        "scale": 1e-1,
        "terms": ["cg-1", "cgr-1", "cg⁻¹", "cgr⁻¹"],
        "followed_by": None,
    },
    "per_dg": {
        "dim": "mass",
        "degree": -1,
        "scale": 1e-2,
        "terms": ["dgr-1", "dg-1", "dgr⁻¹", "dg⁻¹"],
        "followed_by": None,
    },
    "per_g": {
        "dim": "mass",
        "degree": -1,
        "scale": 1e-3,
        "terms": ["gr-1", "g-1", "gr⁻¹", "g⁻¹"],
        "followed_by": None,
    },
    "per_kg": {
        "dim": "mass",
        "degree": -1,
        "scale": 1e-6,
        "terms": ["kgr-1", "kg-1", "kgr⁻¹", "kg⁻¹"],
        "followed_by": None,
    },
    "per_ml": {
        "dim": "length",
        "degree": -3,
        "scale": 1e-0,
        "terms": ["ml-1", "ml⁻¹"],
        "followed_by": None,
    },
    "per_cl": {
        "dim": "length",
        "degree": -3,
        "scale": 1e-1,
        "terms": ["cl-1", "cl⁻¹"],
        "followed_by": None,
    },
    "per_dl": {
        "dim": "length",
        "degree": -3,
        "scale": 1e-2,
        "terms": ["dl-1", "dl⁻¹"],
        "followed_by": None,
    },
    "per_l": {
        "dim": "length",
        "degree": -3,
        "scale": 1e-3,
        "terms": ["l-1", "l⁻¹"],
        "followed_by": None,
    },
    "per_mm3": {
        "dim": "length",
        "degree": -3,
        "scale": 1e3,
        "terms": ["mm-3", "mm⁻³"],
        "followed_by": None,
    },
    "per_cm3": {
        "dim": "length",
        "degree": -3,
        "scale": 1e-0,
        "terms": ["cm-3", "cm⁻³", "cc-1", "cc⁻¹"],
        "followed_by": None,
    },
    "per_dm3": {
        "dim": "length",
        "degree": -3,
        "scale": 1e-3,
        "terms": ["dm-3", "dm⁻³"],
        "followed_by": None,
    },
    "per_m3": {
        "dim": "length",
        "degree": -3,
        "scale": 1e-6,
        "terms": ["m-3", "m⁻³"],
        "followed_by": None,
    },
    "per_mui": {
        "dim": "ui",
        "degree": -1,
        "scale": 1e-0,
        "terms": ["mui-1", "mui⁻¹"],
        "followed_by": None,
    },
    "per_dui": {
        "dim": "ui",
        "degree": -1,
        "scale": 1e-1,
        "terms": ["dui-1", "dui⁻¹"],
        "followed_by": None,
    },
    "per_cui": {
        "dim": "ui",
        "degree": -1,
        "scale": 1e-2,
        "terms": ["cui-1", "cui⁻¹"],
        "followed_by": None,
    },
    "per_ui": {
        "dim": "ui",
        "degree": -1,
        "scale": 1e-3,
        "terms": ["ui-1", "ui⁻¹"],
        "followed_by": None,
    },
    # Surfaces
    "per_µm2": {
        "dim": "length",
        "degree": -2,
        "scale": 1e8,
        "terms": ["µm-2", "µm⁻²"],
        "followed_by": None,
    },
    "per_mm2": {
        "dim": "length",
        "degree": -2,
        "scale": 1e2,
        "terms": ["mm-2", "mm⁻²"],
        "followed_by": None,
    },
    "per_cm2": {
        "dim": "length",
        "degree": -2,
        "scale": 1e-0,
        "terms": ["cm-2", "cm⁻²"],
        "followed_by": None,
    },
    "per_dm2": {
        "dim": "length",
        "degree": -2,
        "scale": 1e-2,
        "terms": ["dm-2", "dm⁻²"],
        "followed_by": None,
    },
    "per_m2": {
        "dim": "length",
        "degree": -2,
        "scale": 1e-4,
        "terms": ["m-2", "m⁻²"],
        "followed_by": None,
    },
}


common_measurements = {
    "weight": {
        "unit": "kg",
        "unitless_patterns": [
            {
                "terms": ["poids", "poid", "pese", "pesant", "pesait", "pesent"],
                "ranges": [
                    {"min": 0, "max": 200, "unit": "kg"},
                    {"min": 200, "unit": "g"},
                ],
            }
        ],
    },
    "size": {
        "unit": "m",
        "unitless_patterns": [
            {
                "terms": [
                    "mesure",
                    "taille",
                    "mesurant",
                    "mesurent",
                    "mesurait",
                    "mesuree",
                    "hauteur",
                    "largeur",
                    "longueur",
                ],
                "ranges": [
                    {"min": 0, "max": 3, "unit": "m"},
                    {"min": 3, "unit": "cm"},
                ],
            }
        ],
    },
    "bmi": {
        "unit": "kg_per_m2",
        "unitless_patterns": [
            {"terms": ["imc", "bmi"], "ranges": [{"unit": "kg_per_m2"}]}
        ],
    },
    "volume": {"unit": "m3", "unitless_patterns": []},
}

unit_divisors = ["/", "par"]

stopwords = ["par", "sur", "de", "a", ",", "et"]

# Should we only make accented patterns and expect the user to use
# `eds.normalizer` component first ?
range_patterns = [
    ("De", "à"),
    ("De", "a"),
    ("de", "à"),
    ("de", "a"),
    ("Entre", "et"),
    ("entre", "et"),
    (None, "à"),
]



from edsnlp.core import registry

from . import patterns
from .measurements import MeasurementsMatcher

DEFAULT_CONFIG = dict(
    measurements=list(patterns.common_measurements.keys()),  # noqa: E501
    units_config=patterns.units_config,
    number_terms=patterns.number_terms,
    number_regex=patterns.number_regex,
    stopwords=patterns.stopwords,
    unit_divisors=patterns.unit_divisors,
    ignore_excluded=True,
    compose_units=True,
    attr="NORM",
    extract_ranges=False,
    range_patterns=patterns.range_patterns,
    after_snippet_limit=6,
    before_snippet_limit=10,
    span_getter=None,
    merge_mode="intersect",
    as_ents=False,
    span_setter=None,
)

create_component = registry.factory.register(
    "eds.measurements",
    assigns=["doc.spans", "doc.ents"],
    deprecated=["eds.measures"],
)(MeasurementsMatcher)



from typing import Any, Dict, Union

from edsnlp.core import PipelineProtocol, registry

from .accents.accents import AccentsConverter
from .normalizer import Normalizer
from .pollution.patterns import default_enabled as default_enabled_pollution
from .pollution.pollution import PollutionTagger
from .quotes.quotes import QuotesConverter
from .spaces.spaces import SpacesTagger

DEFAULT_CONFIG = dict(
    accents=True,
    lowercase=True,
    quotes=True,
    spaces=True,
    pollution=True,
)


@registry.factory.register(
    "eds.normalizer",
    assigns=["token.norm", "token.tag"],
    deprecated=["normalizer"],
)
def create_component(
    nlp: PipelineProtocol,
    name: str = "normalizer",
    *,
    accents: Union[bool, Dict[str, Any]] = True,
    lowercase: Union[bool, Dict[str, Any]] = True,
    quotes: Union[bool, Dict[str, Any]] = True,
    spaces: Union[bool, Dict[str, Any]] = True,
    pollution: Union[bool, Dict[str, Any]] = True,
) -> Normalizer:
    """
    Normalisation pipeline. Modifies the `NORM` attribute,
    acting on five dimensions :

    - `lowercase`: using the default `NORM`
    - `accents`: deterministic and fixed-length normalisation of accents.
    - `quotes`: deterministic and fixed-length normalisation of quotation marks.
    - `spaces`: "removal" of spaces tokens (via the tag_ attribute).
    - `pollution`: "removal" of pollutions (via the tag_ attribute).

    Parameters
    ----------
    nlp: PipelineProtocol
        The pipeline object.
    name : str
        The component name.
    lowercase : bool
        Whether to remove case.
    accents : Union[bool, Dict[str, Any]]
        `Accents` configuration object
    quotes : Union[bool, Dict[str, Any]]
        `Quotes` configuration object
    spaces : Union[bool, Dict[str, Any]]
        `Spaces` configuration object
    pollution : Union[bool, Dict[str, Any]]
        Optional `Pollution` configuration object.
    """

    if accents:
        accents = AccentsConverter(
            nlp=nlp,
            name="eds.accents",
            **(accents if accents is not True else {}),
        )

    if quotes:
        quotes = QuotesConverter(
            nlp=nlp,
            name="eds.quotes",
            **(quotes if quotes is not True else {}),
        )

    if spaces:
        spaces = SpacesTagger(
            nlp=nlp,
            name="eds.spaces",
            **(spaces if spaces is not True else {}),
        )

    if pollution:
        config = dict(default_enabled_pollution)
        if isinstance(pollution, dict):
            pollution = (
                pollution if "pollution" not in pollution else pollution["pollution"]
            )
            config.update(pollution)
        pollution = PollutionTagger(
            nlp=nlp,
            name="eds.pollution",
            pollution=config,
        )

    normalizer = Normalizer(
        nlp=nlp,
        name=name,
        lowercase=lowercase,
        accents=accents or None,
        quotes=quotes or None,
        pollution=pollution or None,
        spaces=spaces or None,
    )

    return normalizer



from __future__ import annotations

import copyreg
import gc
import io
import logging
import math
import multiprocessing
import multiprocessing.reduction
import os
import sys
import tempfile
import warnings
from contextlib import nullcontext
from multiprocessing.connection import wait
from random import shuffle
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
    Union,
)

import dill
from typing_extensions import TypedDict

from edsnlp.core.lazy_collection import LazyCollection
from edsnlp.data.converters import set_current_tokenizer
from edsnlp.utils.batching import batchify_fns, batchify_with_counts
from edsnlp.utils.collections import batchify, flatten

doc_size_fns = {
    "words": len,
}

if TYPE_CHECKING:
    import torch

    from edsnlp.core.torch_component import TorchComponent

Stage = TypedDict(
    "Stage",
    {
        "cpu_components": List[Tuple[str, Callable, Dict, Any]],
        "gpu_component": Optional[Any],
    },
)


def apply_basic_pipes(docs, pipes):
    for name, pipe, kwargs, tok in pipes:
        with set_current_tokenizer(tok):
            if hasattr(pipe, "batch_process"):
                docs = pipe.batch_process(docs)
            else:
                docs = [pipe(doc, **kwargs) for doc in docs]
    return docs


class ForkingPickler(dill.Pickler):
    """
    ForkingPickler that uses dill instead of pickle to transfer objects between
    processes.
    """

    _extra_reducers = {}
    _copyreg_dispatch_table = copyreg.dispatch_table

    def __new__(cls, *args, **kwargs):
        result = dill.Pickler.__new__(ForkingPickler)
        # Python would not call __init__ if the original
        # multiprocessing.reduction.ForkingPickler called, leading to a call to this
        # monkey-patched __new__ method, because [original cls] != [type of result]
        # (see https://docs.python.org/3/reference/datamodel.html#basic-customization)
        # so we force the call to __init__ here
        if not isinstance(result, cls):
            result.__init__(*args, **kwargs)
        return result

    def __init__(self, *args, **kwds):
        super().__init__(*args, **kwds)
        self.dispatch_table = self._copyreg_dispatch_table.copy()
        self.dispatch_table.update(self._extra_reducers)

    @classmethod
    def register(cls, type, reduce):
        """Register a reduce function for a type."""
        cls._extra_reducers[type] = reduce

    @classmethod
    def dumps(cls, obj, protocol=None, *args, **kwds):
        buf = io.BytesIO()
        cls(buf, protocol, *args, **kwds).dump(obj)
        return buf.getbuffer()

    loads = dill.loads


def replace_pickler():
    """
    Replace the default pickler used by multiprocessing with dill.
    "multiprocess" didn't work for obscure reasons (maybe the reducers / dispatchers
    are not propagated between multiprocessing and multiprocess => torch specific
    reducers might be missing ?), so this patches multiprocessing directly.
    directly.

    For some reason I do not explain, this has a massive impact on the performance of
    the multiprocessing backend. With the original pickler, the performance can be
    up to 2x slower than with our custom one.
    """
    old_pickler = multiprocessing.reduction.ForkingPickler

    before = (
        dict(ForkingPickler._extra_reducers),
        old_pickler.__new__,
        old_pickler.dumps,
        old_pickler.loads,
        old_pickler.register,
    )

    old_pickler.__new__ = ForkingPickler.__new__
    old_pickler.dumps = ForkingPickler.dumps
    old_pickler.loads = ForkingPickler.loads
    old_pickler.register = ForkingPickler.register
    ForkingPickler._extra_reducers.update(
        multiprocessing.reduction.ForkingPickler._extra_reducers
    )

    def revert():
        (
            ForkingPickler._extra_reducers,
            old_pickler.__new__,
            old_pickler.dumps,
            old_pickler.loads,
            old_pickler.register,
        ) = before

    return revert


def cpu_count():  # pragma: no cover
    """
    Heavily inspired (partially copied) from joblib's loky
    (https://github.com/joblib/loky/blob/2c21e/loky/backend/context.py#L83)
    by Thomas Moreau and Olivier Grisel.

    Return the number of CPUs we can use to process data in parallel.

    The returned number of CPUs returns the minimum of:
     * `os.cpu_count()`
     * the CPU affinity settings
     * cgroup CPU bandwidth limit (share of total CPU time allowed in a given job)
       typically used in containerized environments like Docker

    Note that on Windows, the returned number of CPUs cannot exceed 61 (or 60 for
    Python < 3.10), see:
    https://bugs.python.org/issue26903.

    It is also always larger or equal to 1.
    """
    # Note: os.cpu_count() is allowed to return None in its docstring
    os_cpu_count = os.cpu_count() or 1
    if sys.platform == "win32":
        # Following loky's windows implementation

        _MAX_WINDOWS_WORKERS = 60
        if sys.version_info >= (3, 8):
            from concurrent.futures.process import _MAX_WINDOWS_WORKERS

            if sys.version_info < (3, 10):
                _MAX_WINDOWS_WORKERS = _MAX_WINDOWS_WORKERS - 1
        os_cpu_count = min(os_cpu_count, _MAX_WINDOWS_WORKERS)

    cpu_count_affinity = os_cpu_count
    try:
        cpu_count_affinity = len(os.sched_getaffinity(0))
    except (NotImplementedError, AttributeError):
        pass

    # Cgroup CPU bandwidth limit available in Linux since 2.6 kernel
    cpu_count_cgroup = os_cpu_count
    cpu_max_fname = "/sys/fs/cgroup/cpu.max"
    cfs_quota_fname = "/sys/fs/cgroup/cpu/cpu.cfs_quota_us"
    cfs_period_fname = "/sys/fs/cgroup/cpu/cpu.cfs_period_us"
    if os.path.exists(cpu_max_fname):
        # cgroup v2
        # https://www.kernel.org/doc/html/latest/admin-guide/cgroup-v2.html
        with open(cpu_max_fname) as fh:
            cpu_quota_us, cpu_period_us = fh.read().strip().split()
    elif os.path.exists(cfs_quota_fname) and os.path.exists(cfs_period_fname):
        # cgroup v1
        # https://www.kernel.org/doc/html/latest/scheduler/sched-bwc.html#management
        with open(cfs_quota_fname) as fh:
            cpu_quota_us = fh.read().strip()
        with open(cfs_period_fname) as fh:
            cpu_period_us = fh.read().strip()
    else:
        cpu_quota_us = "max"
        cpu_period_us = 100_000

    if cpu_quota_us != "max":
        cpu_quota_us = int(cpu_quota_us)
        cpu_period_us = int(cpu_period_us)
        if cpu_quota_us > 0 and cpu_period_us > 0:
            cpu_count_cgroup = math.ceil(cpu_quota_us / cpu_period_us)

    return max(1, min(os_cpu_count, cpu_count_affinity, cpu_count_cgroup))


# Should we check if the multiprocessing module of edsnlp
# is responsible for this child process before replacing the pickler ?
if (
    multiprocessing.current_process() != "MainProcess"
    or hasattr(multiprocessing, "parent_process")
    and multiprocessing.parent_process() is not None
):
    replace_pickler()

DEBUG = True

debug = (
    (lambda *args, flush=False, **kwargs: print(*args, **kwargs, flush=True))
    if DEBUG
    else lambda *args, **kwargs: None
)

if os.environ.get("TORCH_SHARING_STRATEGY"):
    try:
        torch.multiprocessing.set_sharing_strategy(os.environ["TORCH_SHARING_STRATEGY"])
    except NameError:
        pass

try:
    import torch

    # Torch may still be imported as a namespace package, so we can access the
    # torch.save and torch.load functions
    torch_save = torch.save
    torch_load = torch.load

    MAP_LOCATION = None

    try:
        from accelerate.hooks import AlignDevicesHook

        # We need to replace the "execution_device" attribute of the AlignDevicesHook
        # using map_location when unpickling the lazy collection

        def save_align_devices_hook(pickler: Any, obj: Any):
            pickler.save_reduce(load_align_devices_hook, (obj.__dict__,), obj=obj)

        def load_align_devices_hook(state):
            state["execution_device"] = MAP_LOCATION
            new_obj = AlignDevicesHook.__new__(AlignDevicesHook)
            new_obj.__dict__.update(state)
            return new_obj

    except ImportError:
        AlignDevicesHook = None

    def dump(*args, **kwargs):
        # We need to replace the "execution_device" attribute of the AlignDevicesHook
        # using map_location when pickling the lazy collection
        old = None
        try:
            if AlignDevicesHook is not None:
                old = dill.Pickler.dispatch.get(AlignDevicesHook)
                dill.Pickler.dispatch[AlignDevicesHook] = save_align_devices_hook
            dill.settings["recurse"] = True
            return torch_save(*args, pickle_module=dill, **kwargs)
        finally:
            dill.settings["recurse"] = False
            if AlignDevicesHook is not None:
                del dill.Pickler.dispatch[AlignDevicesHook]
                if old is not None:  # pragma: no cover
                    dill.Pickler.dispatch[AlignDevicesHook] = old

    def load(*args, map_location=None, **kwargs):
        global MAP_LOCATION
        MAP_LOCATION = map_location
        if torch.__version__ >= "2.1" and isinstance(args[0], str):
            kwargs["mmap"] = True
        # with open(args[0], "rb") as f:
        #     result = dill.load(f, **kwargs)
        try:
            if torch.__version__ < "2.0.0":
                pickle = torch_load.__globals__["pickle"]
                torch_load.__globals__["pickle"] = dill
            result = torch_load(
                *args,
                pickle_module=dill,
                map_location=map_location,
                **kwargs,
            )
        finally:
            import pickle

            torch_load.__globals__["pickle"] = pickle
        MAP_LOCATION = None
        return result

except (ImportError, AttributeError):  # pragma: no cover

    def load(file, *args, map_location=None, **kwargs):
        # check if path
        if isinstance(file, str):
            with open(file, "rb") as f:
                return dill.load(f, *args, **kwargs)
        return dill.load(file, *args, **kwargs)

    dump = dill.dump


class Exchanger:
    def __init__(
        self,
        mp: multiprocessing.context.BaseContext,
        num_stages: int,
        num_gpu_workers: int,
        num_cpu_workers: int,
        gpu_worker_devices: List[Any],
    ):
        # queue for cpu input tasks
        self.gpu_worker_devices = gpu_worker_devices
        self.num_cpu_workers = num_cpu_workers
        self.num_gpu_workers = num_gpu_workers
        # We add prioritized queue at the end for STOP signals
        self.cpu_inputs_queues = [
            [mp.Queue()] + [mp.SimpleQueue() for _ in range(num_stages + 1)]
            # The input queue is not shared between processes, since calling `wait`
            # on a queue reader from multiple processes may lead to a deadlock
            for _ in range(num_cpu_workers)
        ]
        self.gpu_inputs_queues = [
            [mp.Queue() for _ in range(num_stages + 1)] for _ in range(num_gpu_workers)
        ]
        self.outputs_queue = mp.Queue()
        self.num_stages = num_stages

    # noinspection PyUnresolvedReferences
    def get_cpu_task(self, idx, get_instant_active_or_skip: bool = False):
        queues = self.cpu_inputs_queues[idx]
        if get_instant_active_or_skip:
            # Don't get new tasks
            queues = queues[1:]
        queue_readers = wait(
            [queue._reader for queue in queues],
            timeout=0 if get_instant_active_or_skip else None,
        )
        if len(queue_readers) == 0:
            return None, None
        stage, queue = next(
            (stage, q)
            for stage, q in reversed(list(enumerate(self.cpu_inputs_queues[idx])))
            if q._reader in queue_readers
        )
        item = queue.get()
        return stage, item

    def put_cpu(self, item, stage, idx):
        return self.cpu_inputs_queues[idx][stage].put(item)

    def get_gpu_task(self, idx):
        queue_readers = wait([queue._reader for queue in self.gpu_inputs_queues[idx]])
        stage, queue = next(
            (stage, q)
            for stage, q in reversed(list(enumerate(self.gpu_inputs_queues[idx])))
            if q._reader in queue_readers
        )
        item = queue.get()
        return stage, item

    def put_gpu(self, item, stage, idx):
        return self.gpu_inputs_queues[idx][stage].put(item)

    def put_results(self, items):
        self.outputs_queue.put(items)

    def iter_results(self):
        for out in iter(self.outputs_queue.get, None):
            yield out


class CPUWorker:
    def __init__(
        self,
        cpu_idx: int,
        exchanger: Exchanger,
        gpu_pipe_names: List[str],
        lazy_collection_path: str,
        device: Union[str, "torch.device"],
    ):
        super(CPUWorker, self).__init__()

        self.cpu_idx = cpu_idx
        self.exchanger = exchanger
        self.gpu_pipe_names = gpu_pipe_names
        self.lazy_collection_path = lazy_collection_path
        self.device = device

    def run(self):
        # Cannot pass torch tensor during init i think ? otherwise i get
        # ValueError: bad value(s) in fds_to_keep
        # mp._prctl_pr_set_pdeathsig(signal.SIGINT)
        next_batch_id = self.cpu_idx
        new_batch_iterator = None

        def split_task_into_new_batches(task):
            nonlocal next_batch_id, new_batch_iterator
            task_id, fragments = task
            chunks = list(batchify(lc.reader.read_worker(fragments), lc.chunk_size))
            for chunk_idx, docs in enumerate(chunks):
                # If we sort by size, we must first create the documents
                # to have features against which we will sort
                docs = apply_basic_pipes(docs, preprocess_pipes)

                if lc.sort_chunks:
                    docs.sort(key=doc_size_fns.get(lc.sort_chunks, len))

                batches = [
                    batch
                    for batch in batchify_fns[lc.batch_by](
                        docs,
                        batch_size=lc.batch_size,
                    )
                ]

                for batch_idx, batch in enumerate(batches):
                    assert len(batch) > 0
                    batch_id = next_batch_id

                    # We mark the task id only for the last batch of a task
                    # since the purpose of storing the task id is to know
                    # when the worker has finished processing the task,
                    # which is true only when the last batch has been
                    # processed
                    active_batches[batch_id] = (
                        batch,
                        task_id
                        if (batch_idx == len(batches) - 1)
                        and (chunk_idx == len(chunks) - 1)
                        else None,
                    )
                    next_batch_id += num_cpu
                    # gpu_idx = None
                    # batch_id = we have just created a new batch
                    # result from the last stage = None
                    if batch_idx == len(batches) - 1 and chunk_idx == len(chunks) - 1:
                        new_batch_iterator = None
                    yield 0, (None, batch_id, None)

            new_batch_iterator = None

        def read_tasks():
            nonlocal new_batch_iterator

            expect_new_tasks = True

            while expect_new_tasks or len(active_batches) > 0:
                # Check that there are no more than `chunk_size` docs being processed.
                # If there is still room, we can process new batches
                has_room_for_new_batches = (
                    sum(len(ab[0]) for ab in active_batches.values()) < lc.chunk_size
                )

                # if new_batch_iterator is not None and len(active_batches) == 0:
                #     yield next(new_batch_iterator)
                #     continue

                stage, task = self.exchanger.get_cpu_task(
                    idx=self.cpu_idx,
                    # We don't have to wait for new active batches to come back if:
                    get_instant_active_or_skip=(
                        # - we have room for more batches
                        has_room_for_new_batches
                        # - and the batch iterator is still active
                        and new_batch_iterator is not None
                    ),
                )

                # No active batch was returned, and by construction we have room for
                # new batches, so we can start a new batch
                if stage is None:
                    yield next(new_batch_iterator)
                    continue

                # stage, task = next(iterator)
                # Prioritized STOP signal: something bad happened in another process
                # -> stop listening to input queues and raise StopIteration (return)
                if task is None and stage == self.exchanger.num_stages + 1:
                    return
                # Non prioritized STOP signal: there are no more tasks to process
                # and we should smoothly stop (wait that there are no more active
                # tasks, and finalize the writer)
                if stage == 0 and task is None:
                    expect_new_tasks = False
                    continue

                # If first stage, we receive tasks that may require batching
                # again => we split them into chunks
                if stage == 0:
                    new_batch_iterator = split_task_into_new_batches(task)
                    yield next(new_batch_iterator)
                else:
                    yield stage, task

        try:
            lc: LazyCollection = load(
                self.lazy_collection_path, map_location=self.device
            )
            lc.eval()
            preprocess_pipes = []
            num_cpu = self.exchanger.num_cpu_workers
            split_into_batches_after = lc.split_into_batches_after
            if split_into_batches_after is None and (
                lc.batch_by != "docs" or lc.sort_chunks
            ):
                split_into_batches_after = next(
                    (s[0] for s in lc.pipeline if s[0] == "_ensure_doc"), None
                )
            is_before_split = split_into_batches_after is not None

            stages: List[Stage] = [{"cpu_components": [], "gpu_component": None}]
            for name, pipe, *rest in lc.pipeline:
                if name in self.gpu_pipe_names:
                    is_before_split = False
                    stages[-1]["gpu_component"] = pipe
                    stages.append({"cpu_components": [], "gpu_component": None})
                else:
                    if is_before_split:
                        preprocess_pipes.append((name, pipe, *rest))
                    else:
                        stages[-1]["cpu_components"].append((name, pipe, *rest))
                if name is split_into_batches_after:
                    is_before_split = False

            # Start at cpu_idx to avoid having all workers sending their
            # first batch (0 % num_device, cf below) to the same gpu
            active_batches = {}

            logging.info(f"Starting {self} on {os.getpid()}")

            # Inform the main process that we are ready
            self.exchanger.put_results((None, 0, None, None))

            for stage, (gpu_idx, batch_id, result) in read_tasks():
                docs, task_id = active_batches.pop(batch_id)
                count = len(docs)
                for name, pipe, *rest in lc.pipeline:
                    if hasattr(pipe, "enable_cache"):
                        pipe.enable_cache(batch_id)
                if stage > 0:
                    gpu_pipe = stages[stage - 1]["gpu_component"]
                    docs = (
                        gpu_pipe.postprocess(docs, result)
                        if getattr(gpu_pipe, "postprocess", None) is not None
                        else result
                    )

                docs = apply_basic_pipes(docs, stages[stage]["cpu_components"])

                gpu_pipe: "TorchComponent" = stages[stage]["gpu_component"]
                if gpu_pipe is not None:
                    active_batches[batch_id] = (docs, task_id)
                    if gpu_idx is None:
                        gpu_idx = batch_id % len(self.exchanger.gpu_worker_devices)
                    device = self.exchanger.gpu_worker_devices[gpu_idx]
                    self.exchanger.put_gpu(
                        item=(
                            self.cpu_idx,
                            batch_id,
                            gpu_pipe.prepare_batch(docs, device=device),
                        ),
                        idx=gpu_idx,
                        stage=stage,
                    )
                else:
                    for name, pipe, *rest in lc.pipeline:
                        if hasattr(pipe, "disable_cache"):
                            pipe.disable_cache(batch_id)
                    results, count = (
                        lc.writer.write_worker(docs)
                        if lc.writer is not None
                        else (docs, count)
                    )
                    self.exchanger.put_results(
                        (
                            results,
                            count,
                            self.cpu_idx,
                            task_id,
                        )
                    )

            results, count = lc.writer.finalize() if lc.writer is not None else ([], 0)
            self.exchanger.put_results((results, count, self.cpu_idx, "finalize"))

        except BaseException as e:
            import traceback

            print(f"Error in {self}:\n{traceback.format_exc()}", flush=True)
            self.exchanger.put_results((e, 0, self.cpu_idx, None))
        # We need to drain the queues of GPUWorker fed inputs (pre-moved to GPU)
        # to ensure no tensor allocated on producer processes (CPUWorker via
        # collate) are left in consumer processes
        task = True  # anything but None
        stage = None
        while (stage, task) != (0, None):
            try:
                stage, task = self.exchanger.get_cpu_task(self.cpu_idx)
            finally:
                pass

    def __repr__(self):
        return f"<CPUWorker idx={self.cpu_idx}>"


class GPUWorker:
    def __init__(
        self,
        gpu_idx,
        exchanger: Exchanger,
        gpu_pipe_names: List[str],
        lazy_collection_path: str,
        device: Union[str, "torch.device"],
    ):
        super().__init__()

        self.device = device
        self.gpu_idx = gpu_idx
        self.exchanger = exchanger

        self.gpu_pipe_names = gpu_pipe_names
        self.lazy_collection_path = lazy_collection_path

    def run(self):
        import torch

        # mp._prctl_pr_set_pdeathsig(signal.SIGINT)
        try:
            lc = load(self.lazy_collection_path, map_location=self.device)
            lc.eval()
            stage_components = [
                pipe
                # move_to_device(pipe, self.device)
                for name, pipe, *_ in lc.pipeline
                if name in self.gpu_pipe_names
            ]

            del lc
            logging.info(f"Starting {self} on {os.getpid()}")

            # Inform the main process that we are ready
            self.exchanger.put_results((None, 0, None, None))

            with torch.no_grad():  # , torch.cuda.amp.autocast():
                while True:
                    stage, task = self.exchanger.get_gpu_task(self.gpu_idx)
                    if task is None:
                        break

                    cpu_idx, batch_id, batch = task
                    pipe = stage_components[stage]
                    pipe.enable_cache(batch_id)
                    res = pipe.module_forward(batch)
                    self.exchanger.put_cpu(
                        item=(
                            self.gpu_idx,
                            batch_id,
                            {
                                k: v.to("cpu") if hasattr(v, "to") else v
                                for k, v in res.items()
                            },
                        ),
                        stage=stage + 1,
                        idx=cpu_idx,
                    )
                    if stage == len(stage_components) - 1:
                        pipe.disable_cache(batch_id)
                    del batch, task

                task = batch = res = None  # noqa
        except BaseException as e:
            import traceback

            print(f"Error in {self}:\n{traceback.format_exc()}", flush=True)
            self.exchanger.put_results((e, 0, None, None))

        from edsnlp.core.torch_component import _caches

        task = batch = res = None  # noqa
        _caches.clear()
        gc.collect()
        sys.modules["torch"].cuda.empty_cache()

        # We need to drain the queues of CPUWorker fed inputs (pre-moved to GPU)
        # to ensure no tensor allocated on producer processes (CPUWorker via
        # collate) are left in consumer processes
        stage = None
        task = None
        while (stage, task) != (0, None):
            try:
                stage, task = self.exchanger.get_gpu_task(self.gpu_idx)
            finally:
                pass

    def __repr__(self):
        return f"<GPUWorker idx={self.gpu_idx}>"


def execute_multiprocessing_backend(
    lc: LazyCollection,
):
    """
    If you have multiple CPU cores, and optionally multiple GPUs, we provide the
    `multiprocessing` backend that allows to run the inference on multiple
    processes.

    This accelerator dispatches the batches between multiple workers
    (data-parallelism), and distribute the computation of a given batch on one or two
    workers (model-parallelism):

    - a `CPUWorker` which handles the non deep-learning components and the
      preprocessing, collating and postprocessing of deep-learning components
    - a `GPUWorker` which handles the forward call of the deep-learning components

    If no GPU is available, no `GPUWorker` is started, and the `CPUWorkers` handle
    the forward call of the deep-learning components as well.

    The advantage of dedicating a worker to the deep-learning components is that it
    allows to prepare multiple batches in parallel in multiple `CPUWorker`, and ensure
    that the `GPUWorker` never wait for a batch to be ready.

    The overall architecture described in the following figure, for 3 CPU workers and 2
    GPU workers.

    <div style="text-align:center">
        <img src="/assets/images/multiprocessing.png" style="height:400px" />
    </div>

    Here is how a small pipeline with rule-based components and deep-learning components
    is distributed between the workers:

    <div style="text-align:center">
        <img src="/assets/images/model-parallelism.png" />
    </div>

    !!! warning "Caveat"

        Since workers can produce their results in any order, the order of the results
        may not be the same as the order of the input tasks.

    """
    steps = lc.pipeline
    num_cpu_workers = lc.num_cpu_workers
    num_gpu_workers = lc.num_gpu_workers
    show_progress = lc.show_progress
    process_start_method = lc.process_start_method

    # Infer which pipes should be accelerated on GPU
    gpu_steps_candidates = [
        name
        for name, pipe, *_ in steps
        if hasattr(pipe, "module_forward") and hasattr(pipe, "prepare_batch")
    ]
    gpu_pipe_names = (
        gpu_steps_candidates if lc.gpu_pipe_names is None else lc.gpu_pipe_names
    )
    if set(gpu_pipe_names) - set(gpu_steps_candidates):
        raise ValueError(
            "GPU accelerated pipes {} could not be found in the model".format(
                sorted(set(gpu_pipe_names) - set(gpu_steps_candidates))
            )
        )

    old_environ = {
        k: os.environ.get(k) for k in ("TOKENIZERS_PARALLELISM", "OMP_NUM_THREADS")
    }
    if lc.disable_implicit_parallelism:
        os.environ["TOKENIZERS_PARALLELISM"] = "false"
        os.environ["OMP_NUM_THREADS"] = "1"

    requires_gpu = (
        num_gpu_workers is None
        and len(gpu_pipe_names)
        or num_gpu_workers is not None
        and num_gpu_workers > 0
    )

    num_cpus = cpu_count()
    num_devices = 0
    if requires_gpu:
        import torch

        num_devices = torch.cuda.device_count()
        logging.info(f"Number of available devices: {num_devices}")

        if num_gpu_workers is None:
            num_gpu_workers = min(num_devices, num_cpus // 2)
    else:
        num_gpu_workers = 0

    if "torch" in sys.modules:
        try:
            import torch.multiprocessing

            os.environ[
                "TORCH_SHARING_STRATEGY"
            ] = torch.multiprocessing.get_sharing_strategy()
        except ImportError:  # pragma: no cover
            pass

    if any(gpu_steps_candidates):
        if process_start_method == "fork":
            warnings.warn(
                "Using fork start method with GPU workers may lead to deadlocks. "
                "Consider using process_start_method='spawn' instead."
            )

        process_start_method = process_start_method or "spawn"

    default_method = multiprocessing.get_start_method()
    if process_start_method is not None and default_method != process_start_method:
        logging.info(f"Switching process start method to {process_start_method}")

    mp = multiprocessing.get_context(process_start_method)
    max_cpu_workers = max(num_cpus - num_gpu_workers - 1, 0)
    num_cpu_workers = (
        max_cpu_workers
        if num_cpu_workers is None
        else max_cpu_workers + num_cpu_workers + 1
        if num_cpu_workers < 0
        else num_cpu_workers
    )

    if num_gpu_workers == 0:
        gpu_pipe_names = []

    gpu_worker_devices = (
        [
            f"cuda:{gpu_idx * num_devices // num_gpu_workers}"
            for gpu_idx in range(num_gpu_workers)
        ]
        if requires_gpu and lc.gpu_worker_devices is None
        else []
        if lc.gpu_worker_devices is None
        else lc.gpu_worker_devices
    )
    cpu_worker_devices = (
        ["cpu"] * num_cpu_workers
        if lc.cpu_worker_devices is None
        else lc.cpu_worker_devices
    )
    assert len(cpu_worker_devices) == num_cpu_workers
    assert len(gpu_worker_devices) == num_gpu_workers
    if num_cpu_workers == 0:  # pragma: no cover
        (
            num_cpu_workers,
            num_gpu_workers,
            cpu_worker_devices,
            gpu_worker_devices,
            gpu_pipe_names,
        ) = (num_gpu_workers, 0, gpu_worker_devices, [], [])

    exchanger = Exchanger(
        mp,
        num_stages=len(gpu_pipe_names),
        num_cpu_workers=num_cpu_workers,
        num_gpu_workers=num_gpu_workers,
        gpu_worker_devices=gpu_worker_devices,
    )

    # lc = lc.to("cpu")

    cpu_workers = []
    gpu_workers = []

    with tempfile.NamedTemporaryFile(delete=False) as fp:
        dump(lc.worker_copy(), fp)
        fp.close()

    revert_pickler = replace_pickler()

    for gpu_idx in range(num_gpu_workers):
        gpu_workers.append(
            mp.Process(
                target=GPUWorker.run,
                args=(
                    GPUWorker(
                        gpu_idx=gpu_idx,
                        exchanger=exchanger,
                        gpu_pipe_names=gpu_pipe_names,
                        lazy_collection_path=fp.name,
                        device=gpu_worker_devices[gpu_idx],
                    ),
                ),
            )
        )

    for cpu_idx in range(num_cpu_workers):
        cpu_workers.append(
            mp.Process(
                target=CPUWorker.run,
                args=(
                    CPUWorker(
                        cpu_idx=cpu_idx,
                        exchanger=exchanger,
                        gpu_pipe_names=gpu_pipe_names,
                        lazy_collection_path=fp.name,
                        device=cpu_worker_devices[cpu_idx],
                    ),
                ),
            )
        )

    logging.info(f"Main PID {os.getpid()}")

    logging.info(
        f"Starting {num_cpu_workers} cpu workers and {num_gpu_workers} gpu workers on "
        f"{gpu_worker_devices}, with accelerated pipes: {gpu_pipe_names}",
    )

    for worker in (*cpu_workers, *gpu_workers):
        worker.start()

    logging.info("Workers are ready")

    for i in range(len((*cpu_workers, *gpu_workers))):
        outputs, count, cpu_idx, output_task_id = exchanger.outputs_queue.get()
        if isinstance(outputs, BaseException):
            raise outputs

    os.unlink(fp.name)

    num_max_enqueued = 1
    # Number of input/output batch per process
    outputs_iterator = exchanger.iter_results()

    cpu_worker_indices = list(range(num_cpu_workers))
    inputs_iterator = lc.reader.read_main()
    active_chunks = [{} for i in cpu_worker_indices]
    non_finalized = {i for i in cpu_worker_indices}
    max_workload = lc.chunk_size * num_max_enqueued

    bar = nullcontext()
    if show_progress:
        from tqdm import tqdm

        bar = tqdm(smoothing=0.1, mininterval=1.0)

    def get_and_process_output():
        outputs, count, cpu_idx, output_task_id = next(outputs_iterator)
        if output_task_id == "finalize":
            non_finalized.discard(cpu_idx)
        if isinstance(outputs, BaseException):
            raise outputs
        if show_progress:
            bar.update(count)
        if count > 0:
            yield outputs
        if output_task_id is not None:
            active_chunks[cpu_idx].pop(output_task_id, None)

    def process():
        try:
            with bar, lc.eval():
                for input_task_id, (batch, batch_size) in enumerate(
                    batchify_with_counts(inputs_iterator, lc.chunk_size)
                ):
                    while all(sum(wl.values()) >= max_workload for wl in active_chunks):
                        yield from get_and_process_output()

                    # Shuffle to ensure the first process does not receive all the
                    # documents in case of workload equality
                    shuffle(cpu_worker_indices)
                    cpu_idx = min(
                        cpu_worker_indices,
                        key=lambda i: sum(active_chunks[i].values()),
                    )
                    exchanger.put_cpu((input_task_id, batch), stage=0, idx=cpu_idx)
                    active_chunks[cpu_idx][input_task_id] = batch_size

                # Inform the CPU workers that there are no more tasks to process
                for i, worker in enumerate(cpu_workers):
                    exchanger.cpu_inputs_queues[i][0].put(None)

                while any(active_chunks):
                    yield from get_and_process_output()

                while len(non_finalized):
                    yield from get_and_process_output()

        finally:
            revert_pickler()

            for k, v in old_environ.items():
                os.environ.pop(k, None)
                if v is not None:
                    os.environ[k] = v

            # Send gpu and cpu process the order to stop processing data
            # We use the prioritized queue to ensure the stop signal is processed
            # before the next batch of data
            for i, worker in enumerate(gpu_workers):
                exchanger.gpu_inputs_queues[i][-1].put(None)
            for i, worker in enumerate(cpu_workers):
                exchanger.cpu_inputs_queues[i][-1].put(None)

            # Enqueue a final non prioritized STOP signal to ensure there remains no
            # data in the queues (cf drain loop in CPUWorker / GPUWorker)
            for i, worker in enumerate(gpu_workers):
                exchanger.gpu_inputs_queues[i][0].put(None)
            for i, worker in enumerate(gpu_workers):
                worker.join(timeout=5)
            for i, worker in enumerate(cpu_workers):
                exchanger.cpu_inputs_queues[i][0].put(None)
            for i, worker in enumerate(cpu_workers):
                worker.join(timeout=1)

            # If a worker is still alive, kill it
            # This should not happen, but for a reason I cannot explain, it does in
            # some CPU workers sometimes when we catch an error, even though each run
            # method of the workers completes cleanly. Maybe this has something to do
            # with the cleanup of these processes ?
            for i, worker in enumerate(gpu_workers):  # pragma: no cover
                if worker.is_alive():
                    logging.error(f"Killing <GPUWorker idx={i}>")
                    worker.kill()
            for i, worker in enumerate(cpu_workers):  # pragma: no cover
                if worker.is_alive():
                    logging.error(f"Killing <CPUWorker idx={i}>")
                    worker.kill()

            for queue_group in (
                *exchanger.cpu_inputs_queues,
                *exchanger.gpu_inputs_queues,
                [exchanger.outputs_queue],
            ):
                for queue in queue_group:
                    if hasattr(queue, "cancel_join_thread"):
                        queue.cancel_join_thread()

    gen = process()
    return flatten(gen) if lc.writer is None else lc.writer.write_main(gen)



import pytest
import spacy
from pytest import fixture
from spacy.tokens import Span
from typing_extensions import Literal

import edsnlp
import edsnlp.pipes as eds

if not Span.has_extension("cui"):
    Span.set_extension("cui", default=None)


@fixture
def gold():
    blank_nlp = edsnlp.blank("eds")
    doc1 = blank_nlp.make_doc("Prise de folfox puis doliprane par le patient.")

    doc1.ents = [
        Span(doc1, 0, 1, "event"),
        Span(doc1, 2, 3, "drug"),
        Span(doc1, 4, 5, "drug"),
        Span(doc1, 7, 8, "livb"),
    ]
    doc1.ents[0]._.cui = None
    doc1.ents[1]._.cui = "CONCEPT1"
    doc1.ents[2]._.cui = "CONCEPT2"
    doc1.ents[3]._.cui = "CONCEPT3"

    return [doc1]


@pytest.mark.parametrize(
    "metric,probability_mode,reference_mode",
    [
        ("cosine", "softmax", "concept"),
        ("cosine", "sigmoid", "concept"),
        ("dot", "sigmoid", "synonym"),
    ],
)
def test_span_linker(
    metric: Literal["cosine", "dot"],
    probability_mode: Literal["softmax", "sigmoid"],
    reference_mode: Literal["concept", "synonym"],
    gold,
    tmp_path,
):
    nlp = edsnlp.blank("eds")
    nlp.add_pipe(
        eds.span_linker(
            rescale=20.0,
            threshold=0.8,
            metric=metric,
            reference_mode=reference_mode,
            probability_mode=probability_mode,
            # just to maximize coverage, prefer init_weights=True in practice
            init_weights=True if reference_mode == "concept" else False,
            embedding=eds.span_pooler(
                span_getter=["ents"],
                hidden_size=128,
                embedding=eds.transformer(
                    span_getter=["ents"],
                    model="prajjwal1/bert-tiny",
                    window=128,
                    stride=96,
                ),
            ),
        ),
        name="linker",
    )

    def convert(entry):
        doc = nlp.make_doc(entry["STR"].lower()[:100])
        span = spacy.tokens.Span(
            doc,
            0,
            len(doc),
            label=entry["GRP"],
        )
        span._.cui = entry["CUI"]
        doc.ents = [span]
        return doc

    linker: eds.span_linker = nlp.get_pipe("linker")

    synonyms = edsnlp.data.from_iterable(
        [
            {"STR": "folfox", "CUI": "CONCEPT1", "GRP": "drug"},
            {"STR": "doliprane", "CUI": "CONCEPT2", "GRP": "drug"},
            {"STR": "doliprone", "CUI": "CONCEPT2", "GRP": "drug"},
            {"STR": "patiente", "CUI": "CONCEPT3", "GRP": "livb"},
            {"STR": "docteur", "CUI": "CONCEPT4", "GRP": "livb"},
        ],
        converter=convert,
    )
    nlp.post_init(synonyms)
    pred = nlp.pipe([doc.copy() for doc in gold])

    pred_cuis = [span._.cui for doc in pred for span in doc.ents]
    gold_cuis = [span._.cui for doc in pred for span in doc.ents]

    assert pred_cuis == gold_cuis

    if probability_mode == "softmax":
        batch = linker.prepare_batch([doc.copy() for doc in gold], supervision=True)
        results = linker.forward(batch)
        assert results["loss"] > 0

        nlp.to_disk(tmp_path / "model")
        nlp = edsnlp.load(tmp_path / "model")

    assert "TrainableSpanLinker" in str(linker)



results_hemiplegia = dict(
    has_match=[
        True,
        True,
        True,
    ],
    detailled_status="PRESENT",
    assign=None,
    texts=[
        "Patient hémiplégique",
        "Paralysie des membres inférieurs",
        "Patient en LIS",  # locked-in syndrom
    ],
)



from collections import defaultdict

import torch

from edsnlp.utils.collections import get_deep_attr, set_deep_attr


class ScheduledOptimizer(torch.optim.Optimizer):
    def __init__(self, optim):
        self.optim = optim
        schedule_to_groups = defaultdict(lambda: [])
        for group in self.optim.param_groups:
            if "schedules" in group:
                group["schedules"] = (
                    group["schedules"]
                    if isinstance(group["schedules"], list)
                    else [group["schedules"]]
                )
                group["schedules"] = list(group["schedules"])
                for schedule in group["schedules"]:
                    schedule_to_groups[schedule].append(group)
                    schedule.step(group)

    def zero_grad(self):
        return self.optim.zero_grad()

    @property
    def param_groups(self):
        return self.optim.param_groups

    @param_groups.setter
    def param_groups(self, value):
        self.optim.param_groups = value

    @property
    def state(self):
        return self.optim.state

    @state.setter
    def state(self, value):
        self.optim.state = value

    def state_dict(self):
        state = {
            "optim": self.optim.state_dict(),
            "lr": [group.get("lr") for group in self.optim.param_groups],
            "schedules": [
                [schedule.state_dict() for schedule in group.get("schedules", ())]
                for group in self.optim.param_groups
            ],
        }
        for group in state["optim"]["param_groups"]:
            if "schedules" in group:
                del group["schedules"]
        return state

    def load_state_dict(self, state):
        optim_schedules = [
            group.get("schedules", ()) for group in self.optim.param_groups
        ]
        self.optim.load_state_dict(state["optim"])
        for group, group_schedule, group_schedules_state, lr in zip(
            self.optim.param_groups, optim_schedules, state["schedules"], state["lr"]
        ):
            group["schedules"] = group_schedule
            for schedule, schedule_state in zip(
                group["schedules"], group_schedules_state
            ):
                schedule.load_state_dict(schedule_state)
            group["lr"] = lr

    def step(self, closure=None):
        self.optim.step(closure=closure)
        for group in self.optim.param_groups:
            if "schedules" in group:
                for schedule in group["schedules"]:
                    schedule.step(group)

    def initialize(self):
        self.optim.zero_grad()
        for group in self.optim.param_groups:
            for param in group["params"]:
                if param.requires_grad:
                    param.grad = torch.zeros_like(param)
        self.optim.step()


class OptimizerGroupsProxy:
    def __init__(self, groups):
        self.param_groups = groups


class LinearSchedule:
    def __init__(
        self,
        total_steps,
        max_value=None,
        start_value=0.0,
        path="lr",
        warmup_rate=0.0,
    ):
        self.path = path
        self.start_value = start_value
        self.max_value = max_value
        self.warmup_rate = warmup_rate
        self.total_steps = total_steps
        self.idx = 0

    def state_dict(self):
        return {
            "idx": self.idx,
        }

    def load_state_dict(self, state):
        self.idx = state["idx"]

    def step(self, group, closure=None):
        if self.max_value is None:
            self.max_value = get_deep_attr(group, self.path)
        warmup_steps = self.total_steps * self.warmup_rate
        if self.idx < warmup_steps:
            progress = self.idx / warmup_steps
            value = self.start_value + (self.max_value - self.start_value) * progress
        else:
            progress = (self.idx - warmup_steps) / (self.total_steps - warmup_steps)
            value = self.max_value + (0 - self.max_value) * progress
        self.idx += 1
        set_deep_attr(group, self.path, value)



from typing import List

from pytest import fixture, mark

from edsnlp.pipes.qualifiers.negation import Negation
from edsnlp.utils.examples import parse_example

negation_examples: List[str] = [
    "<ent polarity_=NEG>métas,tases</ent> : non",
    (
        "Pas de <ent negated=true>lésion pulmonaire avec "
        "l'absence de lésion secondaire</ent>."
    ),
    "Cancer non <ent negation=true>métastasé</ent>.",
    "Absence d'<ent negated=true>image osseuse d'allure évolutive</ent>.",
    "il n'y a pas de <ent polarity_=NEG>métas,tases</ent>",
    "Le patient n'est pas <ent polarity_=NEG>malade</ent>.",
    "Aucun <ent polarity_=NEG>traitement</ent>.",
    "Le <ent polarity_=AFF>scan</ent> révèle une grosseur.",
    "il y a des <ent polarity_=AFF>métastases</ent>",
    "aucun doute sur les <ent polarity_=AFF>métastases</ent>",
    "il n'y a pas de <ent polarity_=NEG>métastases</ent>",
    "il n'y a pas d' <ent polarity_=NEG>métastases</ent>",
    "il n'y a pas d'<ent polarity_=NEG>métastases</ent>",
    "il n'y a pas d'amélioration de la <ent negated=false>maladie</ent>",
    "<ent negated=true>maladie écartée",
    "Le patient ne <ent negated=true>fume</ent> pas.",
    "Le patient ne <ent negated=true>fume vraiment vraiment</ent> pas.",
    "Le patient ne <ent negated=false>fume</ent> que des cigares.",
    "Le résultat exclut un <ent negated=true>SMD</ent>",
    "Le résultat ne permet pas d'exclure un <ent negated=false>SMD</ent>",
    "Situation aggravée par une <ent negated=false>neutropénie fébrile</ent>."
    "Patient est traité d'une cure d'<ent negated=false>ALECTINIB</ent> depuis le ...",
]


@fixture
def negation_factory(blank_nlp):
    default_config = dict(
        pseudo=None,
        preceding=None,
        following=None,
        termination=None,
        verbs=None,
        attr="NORM",
        within_ents=False,
        explain=True,
        span_getter="ents",
    )

    def factory(on_ents_only, **kwargs) -> Negation:
        config = dict(**default_config)
        config.update(kwargs)

        return Negation(
            nlp=blank_nlp,
            on_ents_only=on_ents_only,
            **config,
        )

    return factory


@mark.parametrize("on_ents_only", [True, False])
def test_negation(blank_nlp, negation_factory, on_ents_only):
    negation = negation_factory(on_ents_only=on_ents_only)

    for example in negation_examples:
        text, entities = parse_example(example=example)

        doc = blank_nlp(text)
        doc.ents = [
            doc.char_span(ent.start_char, ent.end_char, label="ent") for ent in entities
        ]

        doc = negation(doc)

        for entity, ent in zip(entities, doc.ents):
            for modifier in entity.modifiers:
                assert bool(ent._.negation_cues) == (modifier.value in {True, "NEG"})

                assert (
                    getattr(ent._, modifier.key) == modifier.value
                ), f"{modifier.key} labels don't match."

                if not on_ents_only:
                    for token in ent:
                        assert (
                            getattr(token._, modifier.key) == modifier.value
                        ), f"{modifier.key} labels don't match."


def test_negation_within_ents(blank_nlp, negation_factory):
    negation = negation_factory(on_ents_only=True, within_ents=True)

    examples = [
        "<ent negated=true>Lésion pulmonaire avec absence de lésion secondaire</ent>.",
    ]

    for example in examples:
        text, entities = parse_example(example=example)

        doc = blank_nlp(text)
        doc.ents = [
            doc.char_span(ent.start_char, ent.end_char, label="ent") for ent in entities
        ]

        doc = negation(doc)

        for entity, ent in zip(entities, doc.ents):
            for modifier in entity.modifiers:
                assert bool(ent._.negation_cues) == (modifier.value in {True, "NEG"})

                assert (
                    getattr(ent._, modifier.key) == modifier.value
                ), f"{modifier.key} labels don't match."



results_dementia = dict(
    has_match=[
        True,
        True,
        False,
        True,
    ],
    detailled_status="PRESENT",
    assign=None,
    texts=[
        "D'importants déficits cognitifs",
        "Patient atteint de démence",
        "On retrouve des anti-SLA",  # antibody
        "Une maladie de Charcot",
    ],
)



from __future__ import annotations

import contextlib
import sys
from collections import namedtuple
from functools import wraps
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Container,
    Dict,
    Iterable,
    List,
    Optional,
    Tuple,
    Union,
)

from typing_extensions import Literal

import edsnlp.data

if TYPE_CHECKING:
    import torch

    from edsnlp import Pipeline
    from edsnlp.core.torch_component import TorchComponent
    from edsnlp.data.base import BaseReader, BaseWriter

INFER = type("INFER", (), {"__repr__": lambda self: "INFER"})()


def with_non_default_args(fn: Callable) -> Callable:
    @wraps(fn)
    def wrapper(self, **kwargs):
        return fn(self, **kwargs, _non_default_args=kwargs.keys())

    return wrapper


Batchable = namedtuple("Batchable", ["batch_process"])


class GPUOp:
    def __init__(self, prepare_batch, forward, postprocess):
        self.prepare_batch = prepare_batch
        self.module_forward = forward
        self.postprocess = postprocess

    def batch_process(self, docs):
        res = self.module_forward(self.prepare_batch(docs, None))
        return self.postprocess(docs, res) if self.postprocess is not None else res

    def enable_cache(self, cache_id=None):
        pass

    def disable_cache(self, cache_id=None):
        pass


class MetaLazyCollection(type):
    def __getattr__(self, item):
        if item in edsnlp.data.__all__:
            fn = getattr(edsnlp.data, item)
            setattr(self, item, fn)
            return fn
        raise AttributeError(item)

    def __dir__(self):  # pragma: no cover
        return (*super().__dir__(), *edsnlp.data.__all__)


class LazyCollection(metaclass=MetaLazyCollection):
    def __init__(
        self,
        reader: Optional[BaseReader] = None,
        writer: Optional[BaseWriter] = None,
        pipeline: List[Any] = [],
        config={},
    ):
        self.reader = reader
        self.writer = writer
        self.pipeline: List[Tuple[str, Callable, Dict, Any]] = pipeline
        self.config = config

    @property
    def batch_size(self):
        return self.config.get("batch_size", 1)

    @property
    def batch_by(self):
        return self.config.get("batch_by", "docs")

    @property
    def sort_chunks(self):
        return self.config.get("sort_chunks", False)

    @property
    def split_into_batches_after(self):
        return self.config.get("split_into_batches_after", None)

    @property
    def chunk_size(self):
        return self.config.get("chunk_size", self.config.get("batch_size", 128))

    @property
    def disable_implicit_parallelism(self):
        return self.config.get("disable_implicit_parallelism", True)

    @property
    def num_cpu_workers(self):
        return self.config.get("num_cpu_workers")

    @property
    def num_gpu_workers(self):
        return self.config.get("num_gpu_workers")

    @property
    def gpu_pipe_names(self):
        return self.config.get("gpu_pipe_names")

    @property
    def gpu_worker_devices(self):
        return self.config.get("gpu_worker_devices")

    @property
    def cpu_worker_devices(self):
        return self.config.get("cpu_worker_devices")

    @property
    def backend(self):
        backend = self.config.get("backend")
        return {"mp": "multiprocessing"}.get(backend, backend)

    @property
    def show_progress(self):
        return self.config.get("show_progress", False)

    @property
    def process_start_method(self):
        return self.config.get("process_start_method")

    @with_non_default_args
    def set_processing(
        self,
        batch_size: int = INFER,
        batch_by: Literal["docs", "words", "padded_words"] = "docs",
        chunk_size: int = INFER,
        sort_chunks: bool = False,
        split_into_batches_after: str = INFER,
        num_cpu_workers: Optional[int] = INFER,
        num_gpu_workers: Optional[int] = INFER,
        disable_implicit_parallelism: bool = True,
        backend: Optional[Literal["simple", "multiprocessing", "mp", "spark"]] = INFER,
        gpu_pipe_names: Optional[List[str]] = INFER,
        show_progress: bool = False,
        process_start_method: Optional[Literal["fork", "spawn"]] = INFER,
        gpu_worker_devices: Optional[List[str]] = INFER,
        cpu_worker_devices: Optional[List[str]] = INFER,
        _non_default_args: Iterable[str] = (),
    ) -> "LazyCollection":
        """
        Parameters
        ----------
        batch_size: int
            Number of documents to process at a time in a GPU worker (or in the
            main process if no workers are used).
        batch_by: Literal["docs", "words", "padded_words"]
            How to compute the batch size. Can be "docs" or "words" :

            - "docs" (default) is the number of documents.
            - "words" is the total number of words in the documents.
            - "padded_words" is the total number of words in the documents, including
               padding, assuming the documents are padded to the same length.
        chunk_size: int
            Number of documents to build before splitting into batches. Only used
            with "simple" and "multiprocessing" backends. This is also the number of
            documents that will be passed through the first components of the pipeline
            until a GPU worker is used (then the chunks will be split according to the
            `batch_size` and `batch_by` arguments).

            By default, the chunk size is equal to the batch size, or 128 if the batch
            size is not set.
        sort_chunks: bool
            Whether to sort the documents by size before splitting into batches.
        split_into_batches_after: str
            The name of the component after which to split the documents into batches.
            Only used with "simple" and "multiprocessing" backends.
            By default, the documents are split into batches as soon as the input
            are converted into Doc objects.
        num_cpu_workers: int
            Number of CPU workers. A CPU worker handles the non deep-learning components
            and the preprocessing, collating and postprocessing of deep-learning
            components. If no GPU workers are used, the CPU workers also handle the
            forward call of the deep-learning components.
        num_gpu_workers: Optional[int]
            Number of GPU workers. A GPU worker handles the forward call of the
            deep-learning components. Only used with "multiprocessing" backend.
        disable_implicit_parallelism: bool
            Whether to disable OpenMP and Huggingface tokenizers implicit parallelism in
            multiprocessing mode. Defaults to True.
        gpu_pipe_names: Optional[List[str]]
            List of pipe names to accelerate on a GPUWorker, defaults to all pipes
            that inherit from TorchComponent. Only used with "multiprocessing" backend.
            Inferred from the pipeline if not set.
        backend: Optional[Literal["simple", "multiprocessing", "spark"]]
            The backend to use for parallel processing. If not set, the backend is
            automatically selected based on the input data and the number of workers.

            - "simple" is the default backend and is used when `num_cpu_workers` is 1
                and `num_gpu_workers` is 0.
            - "multiprocessing" is used when `num_cpu_workers` is greater than 1 or
                `num_gpu_workers` is greater than 0.
            - "spark" is used when the input data is a Spark dataframe and the output
                writer is a Spark writer.
        show_progress: Optional[bool]
            Whether to show progress bars (only applicable with "simple" and
            "multiprocessing" backends).
        process_start_method: Optional[Literal["fork", "spawn"]]
            Whether to use "fork" or "spawn" as the start method for the multiprocessing
            backend. The default is "fork" on Unix systems and "spawn" on Windows.

            - "fork" is the default start method on Unix systems and is the fastest
                start method, but it is not available on Windows, can cause issues
                with CUDA and is not safe when using multiple threads.
            - "spawn" is the default start method on Windows and is the safest start
                method, but it is not available on Unix systems and is slower than
                "fork".
        gpu_worker_devices: Optional[List[str]]
            List of GPU devices to use for the GPU workers. Defaults to all available
            devices, one worker per device. Only used with "multiprocessing" backend.
        cpu_worker_devices: Optional[List[str]]
            List of GPU devices to use for the CPU workers. Used for debugging purposes.

        Returns
        -------
        LazyCollection
        """
        kwargs = {k: v for k, v in locals().items() if k in _non_default_args}
        return LazyCollection(
            reader=self.reader,
            writer=self.writer,
            pipeline=self.pipeline,
            config={
                **self.config,
                **{k: v for k, v in kwargs.items() if v is not INFER},
            },
        )

    @classmethod
    def ensure_lazy(cls, data):
        from edsnlp.data.base import IterableReader

        if isinstance(data, cls):
            return data
        return cls(reader=IterableReader(data))

    def map(self, pipe, name: Optional[str] = None, kwargs={}) -> "LazyCollection":
        """
        Maps a callable to the documents.

        Parameters
        ----------
        pipe: Any
            The callable to map to the documents.
        name: Optional[str]
            The name of the pipeline step.
        kwargs: Dict
            The keyword arguments to pass to the callable.

        Returns
        -------
        LazyCollection
        """
        name = name or f"{repr(pipe)}-{id(kwargs)}"
        return LazyCollection(
            reader=self.reader,
            writer=self.writer,
            pipeline=[*self.pipeline, (name, pipe, kwargs, None)],
            config=self.config,
        )

    def map_batches(
        self, pipe, name: Optional[str] = None, kwargs={}
    ) -> "LazyCollection":
        """
        Maps a callable to a batch of documents. The callable should take a list of
        inputs and return a **list** of outputs (not a single output).

        Parameters
        ----------
        pipe: Any
            The callable to map to the documents.
        name: Optional[str]
            The name of the pipeline step.
        kwargs: Dict
            The keyword arguments to pass to the callable.

        Returns
        -------
        LazyCollection
        """
        return self.map(Batchable(pipe), name, kwargs)

    def map_gpu(
        self,
        prepare_batch: Callable[[List, Union[str, torch.device]], Any],
        forward: Callable[[Any], Any],
        postprocess: Optional[Callable[[List, Any], Any]] = None,
        name: Optional[str] = None,
    ) -> "LazyCollection":
        """
        Maps a deep learning operation to a batch of documents, on a GPU worker.

        Parameters
        ----------
        prepare_batch: Callable[[List, Union[str, torch.device]], Any]
            A callable that takes a list of documents and a device and returns a batch
            of tensors (or anything that can be passed to the `forward` callable). This
            will be called on a CPU-bound worker, and may be parallelized.
        forward: Callable[[Any], Any]
            A callable that takes the output of `prepare_batch` and returns the output
            of the deep learning operation. This will be called on a GPU-bound worker.
        postprocess: Optional[Callable[[List, Any], Any]]
            An optional callable that takes the list of documents and the output of the
            deep learning operation, and returns the final output. This will be called
            on the same CPU-bound worker that called the `prepare_batch` function.
        name: Optional[str]
            The name of the pipeline step.

        Returns
        -------
        LazyCollection
        """
        return self.map(GPUOp(prepare_batch, forward, postprocess), name, {})

    def map_pipeline(self, model: Pipeline) -> "LazyCollection":
        new_steps = []
        tokenizer = model.tokenizer
        for name, pipe, kwargs, pipe_tokenizer in self.pipeline:
            new_steps.append((name, pipe, kwargs, pipe_tokenizer or tokenizer))
        new_steps.append(("_ensure_doc", model._ensure_doc, {}, tokenizer))
        for name, pipe in model.pipeline:
            if name not in model._disabled:
                new_steps.append((name, pipe, {}, tokenizer))
        config = (
            {**self.config, "batch_size": model.batch_size}
            if self.batch_size is None
            else self.config
        )
        return LazyCollection(
            reader=self.reader,
            writer=self.writer,
            pipeline=new_steps,
            config=config,
        )

    def write(self, writer: BaseWriter, execute: bool = True) -> Any:
        lc = LazyCollection(
            reader=self.reader,
            writer=writer,
            pipeline=self.pipeline,
            config=self.config,
        )
        return lc.execute() if execute else lc

    def execute(self):
        import edsnlp.processing

        backend = self.backend
        if backend is None:
            try:
                SparkReader = sys.modules.get("edsnlp.data.spark").SparkReader
                SparkWriter = sys.modules.get("edsnlp.data.spark").SparkWriter
            except (KeyError, AttributeError):  # pragma: no cover
                SparkReader = SparkWriter = None
            if (
                SparkReader
                and isinstance(self.reader, SparkReader)
                and SparkWriter
                and (self.writer is None or isinstance(self.writer, SparkWriter))
            ):
                backend = "spark"
            elif (
                self.num_cpu_workers is not None or self.num_gpu_workers is not None
            ) and (
                self.num_cpu_workers is not None
                and self.num_cpu_workers > 1
                or self.num_gpu_workers is not None
                and self.num_gpu_workers > 0
            ):
                backend = "multiprocessing"
            else:
                backend = "simple"
        execute = getattr(edsnlp.processing, f"execute_{backend}_backend")
        return execute(self)

    def __iter__(self):
        return iter(self.execute())

    @contextlib.contextmanager
    def cache(self):
        for name, pipe, *_ in self.pipeline:
            if hasattr(pipe, "enable_cache"):
                pipe.enable_cache()
        yield
        for name, pipe, *_ in self.pipeline:
            if hasattr(pipe, "disable_cache"):
                pipe.disable_cache()

    def torch_components(
        self, disable: Container[str] = ()
    ) -> Iterable[Tuple[str, "TorchComponent"]]:
        """
        Yields components that are PyTorch modules.

        Parameters
        ----------
        disable: Container[str]
            The names of disabled components, which will be skipped.

        Returns
        -------
        Iterable[Tuple[str, 'edsnlp.core.torch_component.TorchComponent']]
        """
        for name, pipe, *_ in self.pipeline:
            if name not in disable and hasattr(pipe, "forward"):
                yield name, pipe

    def to(self, device: Union[str, Optional["torch.device"]] = None):  # noqa F821
        """Moves the pipeline to a given device"""
        for name, pipe, *_ in self.torch_components():
            pipe.to(device)
        return self

    def train(self, mode=True):
        """
        Enables training mode on pytorch modules

        Parameters
        ----------
        mode: bool
            Whether to enable training or not
        """

        class context:
            def __enter__(self):
                pass

            def __exit__(ctx_self, type, value, traceback):
                for name, proc in self.torch_components():
                    proc.train(was_training[name])

        was_training = {name: proc.training for name, proc in self.torch_components()}
        for name, proc in self.torch_components():
            proc.train(mode)

        return context()

    def eval(self):
        """
        Enables evaluation mode on pytorch modules
        """
        return self.train(False)

    def worker_copy(self):
        return LazyCollection(
            reader=self.reader.worker_copy(),
            writer=self.writer,
            pipeline=self.pipeline,
            config=self.config,
        )

    def __dir__(self):  # pragma: no cover
        return (*super().__dir__(), *edsnlp.data.__all__)

    def __getattr__(self, item):
        return getattr(LazyCollection, item).__get__(self)


if TYPE_CHECKING:
    # just to add read/from_* and write/to_* methods to the static type hints
    LazyCollection = edsnlp.data  # noqa: F811



import abc
import re
import unicodedata
from collections import defaultdict
from functools import lru_cache
from itertools import repeat
from typing import Any, Dict, Iterable, List, Optional, Set, Tuple, Union

import regex
from spacy.tokens import Doc, Span
from typing_extensions import Literal, NotRequired, TypedDict

from edsnlp.core import PipelineProtocol
from edsnlp.matchers.phrase import EDSPhraseMatcher
from edsnlp.matchers.regex import RegexMatcher
from edsnlp.matchers.utils import get_text
from edsnlp.pipes.base import (
    BaseNERComponent,
    SpanGetterArg,
    SpanSetterArg,
    get_spans,
    validate_span_getter,
)
from edsnlp.pipes.misc.measurements import patterns
from edsnlp.utils.filter import align_spans, filter_spans, get_span_group

__all__ = ["MeasurementsMatcher"]

AFTER_SNIPPET_LIMIT = 6
BEFORE_SNIPPET_LIMIT = 10


class UnitConfig(TypedDict):
    dim: str
    degree: int
    scale: float
    terms: List[str]
    followed_by: Optional[str] = None


class UnitlessRange(TypedDict):
    min: NotRequired[int]
    max: NotRequired[int]
    unit: str


class UnitlessPatternConfig(TypedDict):
    terms: List[str]
    ranges: List[UnitlessRange]


class UnitlessPatternConfigWithName(TypedDict):
    terms: List[str]
    ranges: List[UnitlessRange]
    name: str


class MsrConfig(TypedDict):
    unit: str
    unitless_patterns: NotRequired[List[UnitlessPatternConfig]]
    name: NotRequired[str]


class Measurement(abc.ABC):
    @abc.abstractmethod
    def __len__(self) -> Iterable["SimpleMeasurement"]:
        """
        Number of items in the measure (only one for SimpleMeasurement)

        Returns
        -------
        Iterable["SimpleMeasurement"]
        """

    @abc.abstractmethod
    def __iter__(self) -> Iterable["SimpleMeasurement"]:
        """
        Iter over items of the measure (only one for SimpleMeasurement)

        Returns
        -------
        Iterable["SimpleMeasurement"]
        """

    @abc.abstractmethod
    def __getitem__(self, item) -> "SimpleMeasurement":
        """
        Access items of the measure (only one for SimpleMeasurement)

        Parameters
        ----------
        item : int

        Returns
        -------
        SimpleMeasurement
        """


class UnitRegistry:
    def __init__(self, config: Dict[str, UnitConfig]):
        self.config = {unicodedata.normalize("NFKC", k): v for k, v in config.items()}
        for unit, unit_config in list(self.config.items()):
            if not unit.startswith("per_") and "per_" + unit not in unit_config:
                self.config["per_" + unit] = {
                    "dim": unit_config["dim"],
                    "degree": -unit_config["degree"],
                    "scale": 1 / unit_config["scale"],
                }

    @lru_cache(maxsize=-1)
    def parse_unit(self, unit: str) -> Tuple[str, float]:
        degrees = defaultdict(lambda: [])
        scale = 1
        for part in regex.split("(?<!per)_", unit):
            unit_config = self.config[unicodedata.normalize("NFKC", part)]
            degrees[unit_config["dim"]].append(unit_config["degree"])
            scale *= unit_config["scale"]
        degrees = {
            k: sum(v) if len(set(v)) > 1 else v[0]
            for k, v in degrees.items()
            if sum(v) != 0
        }
        return str(dict(sorted(degrees.items()))), scale


class SimpleMeasurement(Measurement):
    def __init__(self, value, unit, registry):
        """
        The SimpleMeasurement class contains the value and unit
        for a single non-composite measure

        Parameters
        ----------
        value : float
        unit : str
        """
        super().__init__()
        self.value = value
        self.unit = unit
        self.registry = registry

    def __iter__(self):
        return iter((self,))

    def __getitem__(self, item: int):
        assert isinstance(item, int)
        return [self][item]

    def __str__(self):
        return f"{self.value} {self.unit}"

    def __len__(self):
        return 1

    def __repr__(self):
        return f"Measurement({self.value}, {repr(self.unit)})"

    def __eq__(self, other: Any):
        if isinstance(other, SimpleMeasurement):
            return self.convert_to(other.unit) == other.value
        return False

    def __add__(self, other: "SimpleMeasurement"):
        if other.unit == self.unit:
            return self.__class__(self.value + other.value, self.unit, self.registry)
        return self.__class__(
            self.value + other.convert_to(self.unit), self.unit, self.registry
        )

    def __lt__(self, other: Union["SimpleMeasurement", "RangeMeasurement"]):
        return self.convert_to(other.unit) < min((part.value for part in other))

    def __le__(self, other: Union["SimpleMeasurement", "RangeMeasurement"]):
        return self.convert_to(other.unit) <= min((part.value for part in other))

    def convert_to(self, other_unit):
        self_degrees, self_scale = self.registry.parse_unit(self.unit)
        other_degrees, other_scale = self.registry.parse_unit(other_unit)

        if self_degrees != other_degrees:
            raise AttributeError(
                f"Units {self.unit} and {other_unit} are not homogenous"
            )
        ratio = self_scale / other_scale
        if ratio != 1:
            return self.value * ratio
        return self.value

    def __getattr__(self, other_unit):
        return self.convert_to(other_unit)

    @classmethod
    def verify(cls, ent):
        return True


class RangeMeasurement(Measurement):
    def __init__(self, from_value, to_value, unit, registry):
        super().__init__()
        self.value = (from_value, to_value)
        self.unit = unit
        self.registry = registry

    @classmethod
    def from_measurements(cls, a, b):
        a_value = a.value
        b_value = b.convert_to(a.unit)
        return RangeMeasurement(a_value, b_value, a.unit, a.registry)

    def convert_to(self, other_unit):
        self_degrees, self_scale = self.registry.parse_unit(self.unit)
        other_degrees, other_scale = self.registry.parse_unit(other_unit)

        if self_degrees != other_degrees:
            raise AttributeError(
                f"Units {self.unit} and {other_unit} are not homogenous"
            )
        ratio = self_scale / other_scale
        if ratio == 1:
            return self.value
        from_value = self.value[0] * ratio
        to_value = self.value[1] * ratio
        return (from_value, to_value)

    def __iter__(self):
        yield self[0]
        yield self[1]

    def __len__(self):
        return 2

    def __lt__(self, other: Union[SimpleMeasurement, "RangeMeasurement"]):
        return max(self.convert_to(other.unit)) < min((part.value for part in other))

    def __le__(self, other: Union[SimpleMeasurement, "RangeMeasurement"]):
        return max(self.convert_to(other.unit)) <= max((part.value for part in other))

    def __getattr__(self, other_unit):
        return self.convert_to(other_unit)

    def __eq__(self, other: Any):
        if isinstance(other, RangeMeasurement):
            return self.convert_to(other.unit) == other.value
        return False

    def __str__(self):
        return f"{self.value[0]}-{self.value[1]} {self.unit}"

    def __repr__(self):
        return f"RangeMeasurement({self.value}, {repr(self.unit)})"

    def __getitem__(self, item: int):
        assert isinstance(item, int)
        return SimpleMeasurement(self.value[item], self.unit, self.registry)

    @classmethod
    def verify(cls, ent):
        return True


class MeasurementsMatcher(BaseNERComponent):
    '''
    The `eds.measurements` matcher detects and normalizes numerical measurements
    within a medical document.

    !!! warning

        The ``measurements`` pipeline is still in active development and has not
        been rigorously validated. If you come across a measurement expression that
        goes undetected, please file an issue !

    Scope
    -----
    The `eds.measurements` matcher can extract simple (e.g. `3cm`) measurements.
    It can also detect elliptic enumerations (eg `32, 33 et 34kg`) of measurements
    of the same type and split the measurements accordingly.

    The normalized value can then be accessed via the `span._.{measure_name}` attribute,
    for instance `span._.size` or `span._.weight` and be converted on the fly to a
    desired unit. Like for other components, the `span._.value` extension can also be
    used to access the normalized value for any measurement span.


    The current matcher annotates the following measurements out of the box:

    | Measurement name | Example                |
    |------------------|------------------------|
    | `size`           | `1m50`, `1.50m`        |
    | `weight`         | `12kg`, `1kg300`       |
    | `bmi`            | `BMI: 24`, `24 kg.m-2` |
    | `volume`         | `2 cac`, `8ml`         |


    Examples
    --------
    ```python
    import edsnlp, edsnlp.pipes as eds

    nlp = edsnlp.blank("eds")
    nlp.add_pipe(
        eds.measurements(
            measurements=["size", "weight", "bmi"],
            extract_ranges=True,
        ),
    )

    text = """
    Le patient est admis hier, fait 1m78 pour 76kg.
    Les deux nodules bénins sont larges de 1,2 et 2.4mm.
    BMI: 24.

    Le nodule fait entre 1 et 1.5 cm
    """

    doc = nlp(text)

    measurements = doc.spans["measurements"]

    measurements
    # Out: [1m78, 76kg, 1,2, 2.4mm, 24, entre 1 et 1.5 cm]

    measurements[0]
    # Out: 1m78

    str(measurements[0]._.size), str(measurements[0]._.value)
    # Out: ('1.78 m', '1.78 m')

    measurements[0]._.value.cm
    # Out: 178.0

    measurements[2]
    # Out: 1,2

    str(measurements[2]._.value)
    # Out: '1.2 mm'

    str(measurements[2]._.value.mm)
    # Out: 1.2

    measurements[4]
    # Out: 24

    str(measurements[4]._.value)
    # Out: '24 kg_per_m2'

    str(measurements[4]._.value.kg_per_m2)
    # Out: 24

    str(measurements[5]._.value)
    # Out: 1-1.5 cm
    ```

    To extract all sizes in centimeters, and average range measurements, you can
    use the following snippet:

    ```python
    sizes = [
        sum(item.cm for item in m._.value) / len(m._.value)
        for m in doc.spans["measurements"]
        if m.label_ == "size"
    ]
    sizes
    # Out: [178.0, 0.12, 0.24, 1.25]
    ```

    Customization
    -------------
    You can declare custom measurements by altering the patterns:

    ```python
    import edsnlp, edsnlp.pipes as eds

    nlp = edsnlp.blank("eds")
    nlp.add_pipe(
        eds.measurements(
            measurements={
                "my_custom_surface_measurement": {
                    # This measurement unit is homogenous to square meters
                    "unit": "m2",
                    # Handle cases like "surface: 1.8" (implied m2),
                    # vs "surface: 50" (implied cm2)
                    "unitless_patterns": [
                        {
                            "terms": ["surface", "aire"],
                            "ranges": [
                                {"unit": "m2", "min": 0, "max": 9},
                                {"unit": "cm2", "min": 10, "max": 100},
                            ],
                        }
                    ],
                },
            }
        ),
    )
    ```

    Extensions
    ----------
    The `eds.measurements` pipeline declares its extensions dynamically, depending
    on the `measurements` parameter: each measurement gets its own extension, and
    is assigned to a different span group.

    Parameters
    ----------
    nlp : PipelineProtocol
        The pipeline object
    name : str
        The name of the component.
    measurements : Union[str, List[Union[str, MsrConfig]], Dict[str, MsrConfig]]
        A mapping from measure names to MsrConfig
        Each measure's configuration has the following shape:
        ```{ .python .no-check }
        {
          # the unit (e.g. "kg"),
          "unit": str,
          "unitless_patterns": {
            # preceding trigger terms
            "terms": List[str],
            # unitless ranges -> unit patterns
            "ranges": List[
              {"min": int, "max": int, "unit": str},
              {"min": int, "unit": str},
              ...,
            ],
            ...
          }
        }
        ```
    number_terms: Dict[str, List[str]
        A mapping of numbers to their lexical variants
    stopwords: List[str]
        A list of stopwords that do not matter when placed between a unitless
        trigger
        and a number
    unit_divisors: List[str]
        A list of terms used to divide two units (like: m / s)
    attr : str
        Whether to match on the text ('TEXT') or on the normalized text ('NORM')
    ignore_excluded : bool
        Whether to exclude pollution patterns when matching in the text
    compose_units: bool
        Whether to compose units (like "m/s" or "m.s-1")
    extract_ranges: bool
        Whether to extract ranges (like "entre 1 et 2 cm")
    range_patterns: List[Tuple[str, str]]
        A list of "{FROM} xx {TO} yy" patterns to match range measurements
    after_snippet_limit: int
        Maximum word distance after to link a part of a measurement after its number
    before_snippet_limit: int
        Maximum word distance after to link a part of a measurement before its number
    span_setter: Optional[SpanSetterArg]
        How to set the spans in the document. By default, each measurement will
        be assigned to its own span group (using either the "name" field of the
        config, or the key if you passed a dict), and to the "measurements" group.
    span_getter : SpanGetterArg
        Where to look for measurements in the doc. By default, look in the whole doc.
        You can combine this with the `merge_mode` argument for interesting results.
    merge_mode : Literal["intersect", "align"]
        How to merge matches with the spans from `span_getter`, if given:

        - `intersect`: return only the matches that fall in the `span_getter` spans
        - `align`: if a match overlaps a span from `span_getter` (e.g. a match
          extracted by a machine learning model), return the `span_getter` span
          instead, and assign all the parsed information (`._.date` / `._.duration`)
          to it. Otherwise, don't return the date.

    Authors and citation
    --------------------
    The `eds.measurements` pipeline was developed by AP-HP's Data Science team.
    '''

    # fmt: off
    def __init__(
          self,
          nlp: PipelineProtocol,
          name: str = "measurements",
          *,
          measurements: Union[str, List[Union[str, MsrConfig]], Dict[str, MsrConfig]] = list(patterns.common_measurements.keys()),  # noqa: E501
          units_config: Dict[str, UnitConfig] = patterns.units_config,
          number_terms: Dict[str, List[str]] = patterns.number_terms,
          number_regex: str = patterns.number_regex,
          stopwords: List[str] = patterns.stopwords,
          unit_divisors: List[str] = patterns.unit_divisors,
          ignore_excluded: bool = True,
          compose_units: bool = True,
          attr: str = "NORM",
          extract_ranges: bool = False,
          range_patterns: List[Tuple[Optional[str], Optional[str]]] = patterns.range_patterns,  # noqa: E501
          after_snippet_limit: int = 6,
          before_snippet_limit: int = 10,
          span_getter: Optional[SpanGetterArg] = None,
          merge_mode: Literal["intersect", "align"] = "intersect",
          as_ents: bool = False,
          span_setter: Optional[SpanSetterArg] = None,
    ):
        # fmt: on
        if isinstance(measurements, str):
            measurements = [measurements]
        if isinstance(measurements, (list, tuple)):
            measurements = [
                m
                if isinstance(m, dict)
                else {**patterns.common_measurements[m], "name": m}
                for m in measurements
            ]
        elif isinstance(measurements, dict):
            measurements = [{"name": k, **m} for k, m in measurements.items()]

        self.unit_registry = UnitRegistry(units_config)
        self.unitless_patterns: Dict[str, UnitlessPatternConfigWithName] = {}
        self.unit_part_label_hashes: Set[int] = set()
        self.unitless_label_hashes: Set[int] = set()
        self.unit_followers: Dict[str, str] = {}
        self.measure_names: Dict[str, str] = {}
        self.compose_units = compose_units
        self.extract_ranges = extract_ranges
        self.range_patterns = range_patterns
        self.span_getter = (
            validate_span_getter(span_getter)
            if span_getter is not None
            else None
        )
        self.merge_mode = merge_mode
        self.before_snippet_limit = before_snippet_limit
        self.after_snippet_limit = after_snippet_limit

        # MEASURES
        for m in measurements:
            self.measure_names[self.unit_registry.parse_unit(m["unit"])[0]] = m["name"]

        if span_setter is None:
            span_setter = {
                "ents": as_ents,
                "measurements": True,
                **{
                    name: [name]
                    for name in self.measure_names.values()
                }
            }

        super().__init__(nlp=nlp, name=name, span_setter=span_setter)

        self.hash_strings = {
            "stopword": nlp.vocab.strings["stopword"],
            "unitless_stopword": nlp.vocab.strings["unitless_stopword"],
            "number": nlp.vocab.strings["number"],
        }
        self.regex_matcher = RegexMatcher(
            attr=attr,
            ignore_excluded=True,
        )
        self.term_matcher = EDSPhraseMatcher(
            nlp.vocab,
            attr=attr,
            ignore_excluded=ignore_excluded,
            ignore_space_tokens=True,
        )
        for measure_config in measurements:
            name = measure_config["name"]
            unit = measure_config["unit"]
            self.measure_names[self.unit_registry.parse_unit(unit)[0]] = name
            if "unitless_patterns" in measure_config:
                for pattern in measure_config["unitless_patterns"]:
                    pattern_name = f"unitless_{len(self.unitless_patterns)}"
                    self.term_matcher.build_patterns(
                        nlp,
                        terms={
                            pattern_name: pattern["terms"],
                        },
                    )
                    self.unitless_label_hashes.add(nlp.vocab.strings[pattern_name])
                    self.unitless_patterns[pattern_name] = {"name": name, **pattern}

        # NUMBER PATTERNS
        self.regex_matcher.add("number",[number_regex])
        self.number_label_hashes = {nlp.vocab.strings["number"]}
        for number, terms in number_terms.items():
            self.term_matcher.build_patterns(nlp, {number: terms})
            self.number_label_hashes.add(nlp.vocab.strings[number])

        # UNIT PATTERNS
        for unit_name, unit_config in units_config.items():
            self.term_matcher.build_patterns(nlp, {unit_name: unit_config["terms"]})
            if unit_config.get("followed_by") is not None:
                self.unit_followers[unit_name] = unit_config["followed_by"]
            self.unit_part_label_hashes.add(nlp.vocab.strings[unit_name])

        self.unit_part_label_hashes.add(nlp.vocab.strings["per"])
        self.term_matcher.build_patterns(
            nlp,
            {
                "per": unit_divisors,
                "stopword": stopwords,
                "unitless_stopword": [":"],
            },
        )

    def set_extensions(self) -> None:
        """
        Set extensions for the measurements pipeline.
        """
        super().set_extensions()

        for name in self.measure_names.values():
            if not Span.has_extension(name):
                Span.set_extension(name, default=None)

    def extract_units(self, term_matches: Iterable[Span]) -> Iterable[Span]:
        """
        Extracts unit spans from the document by extracting unit atoms (declared in the
        units_config parameter) and aggregating them automatically
        Ex: "il faut 2 g par jour"
        => we extract [g]=unit(g), [par]=divisor(per), [jour]=unit(day)
        => we aggregate these adjacent matches together to compose a new unit g_per_day


        Parameters
        ----------
        term_matches: Iterable[Span]

        Returns
        -------
        Iterable[Span]
        """
        last = None
        units = []
        current = []
        unit_label_hashes = set()
        for unit_part in filter_spans(term_matches):
            if unit_part.label not in self.unit_part_label_hashes:
                continue
            if last is not None and (
                  (
                        unit_part.doc[last.end: unit_part.start].text.strip() != ""
                        and len(current)
                  )
                  or (
                        not self.compose_units
                        and len(current)
                        and current[-1].label_ != "per"
                  )
            ):
                doc = current[0].doc
                # Last non "per" match: we don't want our units to be like `g_per`
                end = next(
                    (i for i, e in list(enumerate(current))[::-1] if e.label_ != "per"),
                    None,
                )
                if end is not None:
                    unit = "_".join(part.label_ for part in current[: end + 1])
                    units.append(Span(doc, current[0].start, current[end].end, unit))
                    unit_label_hashes.add(units[-1].label)
                current = []
                last = None
            current.append(unit_part)
            last = unit_part

        end = next(
            (i for i, e in list(enumerate(current))[::-1] if e.label_ != "per"), None
        )
        if end is not None:
            doc = current[0].doc
            unit = "_".join(part.label_ for part in current[: end + 1])
            units.append(Span(doc, current[0].start, current[end].end, unit))
            unit_label_hashes.add(units[-1].label)

        return units

    @classmethod
    def make_pseudo_sentence(
          cls,
          doclike: Union[Doc, Span],
          matches: List[Tuple[Span, bool]],
          pseudo_mapping: Dict[int, str],
    ) -> Tuple[str, List[int]]:
        """
        Creates a pseudo sentence (one letter per entity)
        to extract higher order patterns
        Ex: the sentence
        "Il font {1}{,} {2} {et} {3} {cm} de long{.}" is transformed into "wn,n,nuw."

        Parameters
        ----------
        doclike: Union[Doc, Span]
            The document or span to transform
        matches: List[(Span, bool)]
            List of tuple of span and whether the span represents a sentence end
        pseudo_mapping: Dict[int, str]
            A mapping from label to char in the pseudo sentence

        Returns
        -------
        (str, List[int])
            - the pseudo sentence
            - a list of offsets to convert match indices into pseudo sent char indices
        """
        pseudo = []
        snippet = doclike if isinstance(doclike, Span) else doclike[:]
        last = snippet.start
        offsets = []
        for ent, is_sent_split in matches:
            if (
                  ent.start != last
                  and not doclike.doc[last: ent.start].text.strip() == ""
            ):
                pseudo.append("w")
            offsets.append(len(pseudo))
            pseudo.append(pseudo_mapping.get(ent.label, "." if is_sent_split else "w"))
            last = ent.end
        if snippet.end != last and doclike.doc[last: snippet.end].text.strip() == "":
            pseudo.append("w")
        pseudo = "".join(pseudo)

        return pseudo, offsets

    def get_matches(self, doc):
        """
        Extract and filter regex and phrase matches in the document
        to prepare the measurement extraction.
        Returns the matches and a list of hashes to quickly find unit matches

        Parameters
        ----------
        doc: Doc

        Returns
        -------
        Tuple[List[(Span, bool)], Set[int]]
            - List of tuples of spans and whether the spans represents a sentence end
            - List of hash label to distinguish unit from other matches
        """
        sent_ends = [doc[i: i + 1] for i in range(len(doc)) if doc[i].is_sent_end]

        regex_matches = list(self.regex_matcher(doc, as_spans=True))
        term_matches = list(self.term_matcher(doc, as_spans=True))

        # Detect unit parts and compose them into units
        units = self.extract_units(term_matches)
        unit_label_hashes = {unit.label for unit in units}

        # Filter matches to prevent matches over dates or doc entities
        non_unit_terms = [
            term
            for term in term_matches
            if term.label not in self.unit_part_label_hashes
        ]

        # Filter out measurement-related spans that overlap already matched
        # entities (in doc.ents or doc.spans["dates"])
        # Note: we also include sentence ends tokens as 1-token spans in those matches
        # Prevent from matching over ents that are not measurement related
        ents = (e for e in doc.ents if e.label_ not in self.measure_names.values())
        spans__keep__is_sent_end = filter_spans(
            [
                # Tuples (span, keep = is measurement related, is sentence end)
                *zip(get_span_group(doc, "dates"), repeat(False), repeat(False)),
                *zip(regex_matches, repeat(True), repeat(False)),
                *zip(non_unit_terms, repeat(True), repeat(False)),
                *zip(units, repeat(True), repeat(False)),
                *zip(ents, repeat(False), repeat(False)),
                *zip(sent_ends, repeat(True), repeat(True)),
            ]
        )

        # Remove non-measurement related spans (keep = False) and sort the matches
        matches_and_is_sentence_end: List[(Span, bool)] = sorted(
            [
                (span, is_sent_end)
                for span, keep, is_sent_end in spans__keep__is_sent_end
                # and remove entities that are not relevant to this pipeline
                if keep
            ]
        )

        return matches_and_is_sentence_end, unit_label_hashes

    def extract_measurements(self, doclike: Doc):
        """
        Extracts measure entities from the document

        Parameters
        ----------
        doclike: Doc

        Returns
        -------
        List[Span]
        """
        doc = doclike.doc if isinstance(doclike, Span) else doclike
        matches, unit_label_hashes = self.get_matches(doclike)

        # Make match slice function to query them
        def get_matches_after(i):
            anchor = matches[i][0]
            for j, (ent, is_sent_end) in enumerate(matches[i + 1:]):
                if not is_sent_end and ent.start > anchor.end + AFTER_SNIPPET_LIMIT:
                    return
                yield j + i + 1, ent

        def get_matches_before(i):
            anchor = matches[i][0]
            for j, (ent, is_sent_end) in enumerate(matches[i::-1]):
                if not is_sent_end and ent.end < anchor.start - BEFORE_SNIPPET_LIMIT:
                    return
                yield i - j, ent

        # Make a pseudo sentence to query higher order patterns in the main loop
        # `offsets` is a mapping from matches indices (ie match n°i) to
        # char indices in the pseudo sentence
        pseudo, offsets = self.make_pseudo_sentence(
            doclike,
            matches,
            {
                self.hash_strings["stopword"]: ",",
                self.hash_strings["unitless_stopword"]: ":",
                self.hash_strings["number"]: "n",
                **{name: "u" for name in unit_label_hashes},
                **{name: "n" for name in self.number_label_hashes},
            },
        )

        measurements = []
        matched_unit_indices = set()
        matched_number_indices = set()

        # Iterate through the number matches
        for number_idx, (number, is_sent_split) in enumerate(matches):
            if not is_sent_split and number.label not in self.number_label_hashes:
                continue

            # Detect the measure value
            try:
                if number.label_ == "number":
                    number_text = (
                        number.text.replace(" ", "")
                        .replace(",", ".")
                        .replace(" ", "")
                        .lstrip("0")
                    )
                    value = eval(number_text or "0")
                else:
                    value = eval(number.label_)
            except (ValueError, SyntaxError):
                continue

            unit_idx = unit_text = unit_norm = None

            # Find the closest unit after the number
            try:
                unit_idx, unit_text = next(
                    (j, ent)
                    for j, ent in get_matches_after(number_idx)
                    if ent.label in unit_label_hashes
                )
                unit_norm = unit_text.label_
            except (AttributeError, StopIteration):
                pass

            # Try to pair the number with this next unit if the two are only separated
            # by numbers and separators alternatively (as in [1][,] [2] [and] [3] cm)
            try:
                pseudo_sent = pseudo[offsets[number_idx] + 1: offsets[unit_idx]]
                if not re.fullmatch(r"(,n)*", pseudo_sent):
                    unit_text, unit_norm = None, None
            except TypeError:
                pass

            # Otherwise, try to infer the unit from the preceding unit to handle cases
            # like (1 meter 50)
            if unit_norm is None and number_idx - 1 in matched_unit_indices:
                try:
                    unit_before = matches[number_idx - 1][0]
                    if unit_before.end == number.start:
                        unit_norm = self.unit_followers[unit_before.label_]
                except (KeyError, AttributeError, IndexError):
                    pass

            # If no unit was matched, try to detect unitless patterns before
            # the number to handle cases like ("Weight: 63, Height: 170")
            if not unit_norm:
                try:
                    (unitless_idx, unitless_text) = next(
                        (j, e)
                        for j, e in get_matches_before(number_idx)
                        if e.label in self.unitless_label_hashes
                    )
                    unit_norm = None
                    if re.fullmatch(
                          r"[,:n]*",
                          pseudo[offsets[unitless_idx] + 1: offsets[number_idx]],
                    ):
                        unitless_pattern = self.unitless_patterns[unitless_text.label_]
                        unit_norm = next(
                            scope["unit"]
                            for scope in unitless_pattern["ranges"]
                            if ("min" not in scope or value >= scope["min"])
                            and ("max" not in scope or value < scope["max"])
                        )
                except StopIteration:
                    pass

            # Otherwise, skip this number
            if not unit_norm:
                continue

            # Compute the final entity
            # TODO: handle this part better without .text.strip(), with cases for
            #  stopwords, etc
            if (
                  unit_text
                  and number.start <= unit_text.end
                  and doc[number.end: unit_text.start].text.strip() == ""
            ):
                ent = doc[number.start: unit_text.end]
            elif (
                  unit_text
                  and unit_text.start <= number.end
                  and doc[unit_text.end: number.start].text.strip() == ""
            ):
                ent = doc[unit_text.start: number.end]
            else:
                ent = number

            # Compute the dimensionality of the parsed unit
            try:
                dims = self.unit_registry.parse_unit(unit_norm)[0]
            except KeyError:
                continue

            # If the measure was not requested, dismiss it
            # Otherwise, relabel the entity and create the value attribute
            if dims not in self.measure_names:
                continue

            ent.label_ = self.measure_names[dims]
            ent._.set(
                self.measure_names[dims],
                SimpleMeasurement(value, unit_norm, self.unit_registry)
            )

            measurements.append(ent)

            if unit_idx is not None:
                matched_unit_indices.add(unit_idx)
            if number_idx is not None:
                matched_number_indices.add(number_idx)

        unmatched = []
        for idx, (match, _) in enumerate(matches):
            if (
                  match.label in unit_label_hashes
                  and idx not in matched_unit_indices
                  or match.label in self.number_label_hashes
                  and idx not in matched_number_indices
            ):
                unmatched.append(match)

        return measurements, unmatched

    @classmethod
    def merge_adjacent_measurements(cls, measurements: List[Span]) -> List[Span]:
        """
        Aggregates extracted measurements together when they are adjacent to handle
        cases like
        - 1 meter 50 cm
        - 30° 4' 54"

        Parameters
        ----------
        measurements: List[Span]

        Returns
        -------
        List[Span]
        """
        merged = measurements[:1]
        for ent in measurements[1:]:
            last = merged[-1]

            if last.end == ent.start and last._.value.unit != ent._.value.unit:
                try:
                    new_value = last._.value + ent._.value
                    merged[-1] = last = last.doc[last.start: ent.end]
                    last.label_ = ent.label_
                    last._.set(last.label_, new_value)
                except (AttributeError, TypeError):
                    merged.append(ent)
            else:
                merged.append(ent)

        return merged

    def merge_measurements_in_ranges(self, measurements: List[Span]) -> List[Span]:
        """
        Aggregates extracted measurements together when they are adjacent to handle
        cases like
        - 1 meter 50 cm
        - 30° 4' 54"

        Parameters
        ----------
        measurements: List[Span]

        Returns
        -------
        List[Span]
        """
        if not self.extract_ranges or not self.range_patterns:
            return measurements

        merged = measurements[:1]
        for ent in measurements[1:]:
            last = merged[-1]

            from_text = last.doc[last.start - 1].norm_ if last.start > 0 else None
            to_text = get_text(last.doc[last.end: ent.start], "NORM", True)
            matching_patterns = [
                (a, b)
                for a, b in self.range_patterns
                if b == to_text and (a is None or a == from_text)
            ]
            if len(matching_patterns):
                try:
                    new_value = RangeMeasurement.from_measurements(
                        last._.value, ent._.value
                    )
                    merged[-1] = last = last.doc[
                                        last.start
                                        if matching_patterns[0][0] is None
                                        else last.start - 1: ent.end
                                        ]
                    last.label_ = ent.label_
                    last._.set(last.label_, new_value)
                except (AttributeError, TypeError):
                    merged.append(ent)
            else:
                merged.append(ent)

        return merged

    def merge_with_existing(
          self,
          extracted: List[Span],
          existing: List[Span],
    ) -> List[Span]:
        """
        Merges the extracted measurements with the existing measurements in the
        document.

        Parameters
        ----------
        extracted: List[Span]
            The extracted measurements
        existing: List[Span]
            The existing measurements in the document

        Returns
        -------
        List[Span]
        """
        if self.merge_mode == "align":
            spans_measurements = align_spans(extracted, existing, sort_by_overlap=True)

            extracted = []
            for span, span_measurements in zip(existing, spans_measurements):
                if len(span_measurements):
                    span._.set(span.label_, span_measurements[0]._.get(span.label_))
                    extracted.append(span)

        elif self.merge_mode == "intersect":
            spans_measurements = align_spans(extracted, existing)
            extracted = []
            for span, span_measurements in zip(existing, spans_measurements):
                extracted.extend(span_measurements)
            extracted = list(dict.fromkeys(extracted))

        return extracted

    def __call__(self, doc):
        """
        Adds measurements to document's "measurements" SpanGroup.

        Parameters
        ----------
        doc:
            spaCy Doc object

        Returns
        -------
        doc:
            spaCy Doc object, annotated for extracted measurements.
        """
        existing = (
            list(get_spans(doc, self.span_getter))
            if self.span_getter is not None
            else ())
        snippets = (
            dict.fromkeys(ent.sent for ent in existing)
            if self.span_getter is not None
            else [doc]
        )
        measurements = [m for s in snippets for m in self.extract_measurements(s)[0]]
        measurements = self.merge_adjacent_measurements(measurements)
        measurements = self.merge_measurements_in_ranges(measurements)
        if self.span_getter is not None:
            measurements = self.merge_with_existing(measurements, existing)

        self.set_spans(doc, measurements)

        return doc



import re
from typing import Dict, List, Optional, Union

from spacy.tokens import Doc, Span

from edsnlp.core import PipelineProtocol
from edsnlp.matchers.regex import RegexMatcher
from edsnlp.pipes.base import BaseComponent
from edsnlp.utils.filter import filter_spans

from . import patterns
from .patterns import default_enabled


class PollutionTagger(BaseComponent):
    """
    Tags pollution tokens.

    Populates a number of spaCy extensions :

    - `Token._.pollution` : indicates whether the token is a pollution
    - `Doc._.clean` : lists non-pollution tokens
    - `Doc._.clean_` : original text with pollutions removed.
    - `Doc._.char_clean_span` : method to create a Span using character
      indices extracted using the cleaned text.

    Parameters
    ----------
    nlp : PipelineProtocol
        The pipeline object
    name : Optional[str]
        The component name.
    pollution : Dict[str, Union[str, List[str]]]
        Dictionary containing regular expressions of pollution.
    """

    # noinspection PyProtectedMember
    def __init__(
        self,
        nlp: PipelineProtocol,
        name: Optional[str] = "pollution",
        *,
        pollution: Dict[str, Union[bool, str, List[str]]] = default_enabled,
    ):
        super().__init__(nlp=nlp, name=name)
        nlp.vocab.strings.add("EXCLUDED")

        self.pollution = dict()

        for k, v in pollution.items():
            if v is True:
                pattern = patterns.pollution[k]
                self.pollution[k] = pattern if isinstance(pattern, list) else [pattern]
            elif isinstance(v, str):
                self.pollution[k] = [v]
            elif isinstance(v, list):
                self.pollution[k] = v

        self.regex_matcher = RegexMatcher(flags=re.MULTILINE)
        self.build_patterns()

    def build_patterns(self) -> None:
        """
        Builds the patterns for phrase matching.
        """

        # efficiently build spaCy matcher patterns
        for k, v in self.pollution.items():
            self.regex_matcher.add(k, v)

    def process(self, doc: Doc) -> List[Span]:
        """
        Find pollutions in doc and clean candidate negations to remove pseudo negations

        Parameters
        ----------
        doc:
            spaCy Doc object

        Returns
        -------
        pollution:
            list of pollution spans
        """

        pollutions = self.regex_matcher(doc, as_spans=True)
        pollutions = filter_spans(pollutions)

        return pollutions

    def __call__(self, doc: Doc) -> Doc:
        """
        Tags pollutions.

        Parameters
        ----------
        doc:
            spaCy Doc object

        Returns
        -------
        doc:
            spaCy Doc object, annotated for pollutions.
        """
        excluded_hash = doc.vocab.strings["EXCLUDED"]
        pollutions = self.process(doc)

        for pollution in pollutions:
            for token in pollution:
                token._.excluded = True
                token.tag = excluded_hash

        doc.spans["pollutions"] = pollutions

        return doc



import edsnlp
from edsnlp.utils.span_getters import make_span_context_getter


def test_span_sentence_getter(lang):
    nlp = edsnlp.blank("eds")
    nlp.add_pipe("eds.normalizer")
    nlp.add_pipe("eds.sentences")
    nlp.add_pipe("eds.matcher", config={"terms": {"sentence": "sentence"}})
    doc = nlp(
        "This is a sentence. "
        "This is another sentence. "
        "This is a third one. "
        "Last sentence."
    )

    span_getter = make_span_context_getter(
        span_getter=["ents"],
        context_words=2,
        overlap_policy="merge",
    )
    spans = span_getter(doc)
    assert [s.text for s in spans] == [
        "This is a sentence. This is another sentence. This",
        ". Last sentence.",
    ]

    span_getter = make_span_context_getter(
        span_getter=["ents"],
        context_words=2,
        overlap_policy="filter",
    )
    spans = span_getter(doc)
    assert [s.text for s in spans] == [
        "This is a sentence. This",
        ". Last sentence.",
    ]

    span_getter = make_span_context_getter(
        span_getter=["ents"],
        context_words=0,
        context_sents=1,
        overlap_policy="filter",
    )
    spans = span_getter(doc)
    assert [s.text for s in spans] == [
        "This is a sentence.",
        "This is another sentence.",
        "Last sentence.",
    ]

    span_getter = make_span_context_getter(
        span_getter=["ents"],
        context_words=0,
        context_sents=2,
        overlap_policy="merge",
    )
    spans = span_getter(doc)
    assert [s.text for s in spans] == [
        (
            "This is a sentence. This is another sentence. "
            "This is a third one. Last sentence."
        )
    ]



from typing import Optional

from pydantic import BaseModel


class AdicapCode(BaseModel):
    code: str
    sampling_mode: Optional[str] = None
    technic: Optional[str] = None
    organ: Optional[str] = None
    pathology: Optional[str] = None
    pathology_type: Optional[str] = None
    behaviour_type: Optional[str] = None

    def norm(self) -> str:
        return self.code

    def __str__(self):
        return self.norm()



from typing import TYPE_CHECKING
from edsnlp.utils.lazy_module import lazify

lazify()

if TYPE_CHECKING:
    from .base import from_iterable, to_iterable
    from .standoff import read_standoff, write_standoff
    from .brat import read_brat, write_brat
    from .json import read_json, write_json
    from .parquet import read_parquet, write_parquet
    from .spark import from_spark, to_spark
    from .pandas import from_pandas, to_pandas
    from .polars import from_polars, to_polars
    from .converters import get_dict2doc_converter, get_doc2dict_converter



import ast
import importlib
import inspect
import logging
import sys
from typing import Union

import astunparse
from griffe import Extension, Object, ObjectNode
from griffe.docstrings.dataclasses import DocstringSectionParameters
from griffe.expressions import Expr
from griffe.logger import patch_loggers


def get_logger(name):
    new_logger = logging.getLogger(name)
    new_logger.setLevel("ERROR")
    return new_logger


patch_loggers(get_logger)

logger = get_logger(__name__)


class EDSNLPDocstrings(Extension):
    def __init__(self):
        super().__init__()

        self.PIPE_OBJ = {}
        self.FACT_MEM = {}
        self.PIPE_TO_FACT = {}

    def on_instance(self, node: Union[ast.AST, ObjectNode], obj: Object) -> None:
        if (
            isinstance(node, ast.Assign)
            and obj.name == "create_component"
            and isinstance(node.value, ast.Call)
            and isinstance(node.value.func, ast.Call)
        ):
            module_name = obj.path.rsplit(".", 1)[0]
            for name, mod in list(sys.modules.items()):
                if name.startswith("edspdf"):
                    importlib.reload(mod)
            module = importlib.reload(importlib.import_module(module_name))

            config_node = node.value.func
            config_node = next(
                (kw.value for kw in config_node.keywords if kw.arg == "default_config"),
                None,
            )
            try:
                default_config = eval(astunparse.unparse(config_node), module.__dict__)
            except Exception:
                default_config = {}

            # import object to get its evaluated docstring
            try:
                runtime_obj = getattr(module, obj.name)
                source = inspect.getsource(runtime_obj)
                self.visit(ast.parse(source))
            except ImportError:
                logger.debug(f"Could not get dynamic docstring for {obj.path}")
                return
            except AttributeError:
                logger.debug(f"Object {obj.path} does not have a __doc__ attribute")
                return

            callee = (
                runtime_obj.__init__
                if hasattr(runtime_obj, "__init__")
                else runtime_obj
            )
            spec = inspect.getfullargspec(callee)
            func_defaults = dict(
                zip(spec.args[-len(callee.__defaults__) :], callee.__defaults__)
                if callee.__defaults__
                else (),
                **(callee.__kwdefaults__ or {}),
            )
            defaults = {**func_defaults, **default_config}
            self.FACT_MEM[obj.path] = (node, obj, defaults)
            pipe_path = runtime_obj.__module__ + "." + runtime_obj.__name__
            self.PIPE_TO_FACT[pipe_path] = obj.path

            if pipe_path in self.PIPE_OBJ:
                pipe = self.PIPE_OBJ[pipe_path]
                obj.docstring = pipe.docstring
            else:
                return
        elif obj.is_class or obj.is_function:
            self.PIPE_OBJ[obj.path] = obj
            if obj.path in self.PIPE_TO_FACT:
                node, fact_obj, defaults = self.FACT_MEM[self.PIPE_TO_FACT[obj.path]]
                fact_obj.docstring = obj.docstring
                obj = fact_obj
            else:
                return
        else:
            return

        if obj.docstring is None:
            return

        param_section: DocstringSectionParameters = None
        obj.docstring.parser = "numpy"
        for section in obj.docstring.parsed:
            if isinstance(section, DocstringSectionParameters):
                param_section = section  # type: ignore

        if param_section is None:
            return

        for param in param_section.value:
            if param.name in defaults:
                param.default = str(defaults[param.name])
            if isinstance(param.default, Expr):
                continue
            if param.default is not None and len(param.default) > 50:
                param.default = param.default[: 50 - 3] + "..."



results_lymphoma = dict(
    has_match=[
        True,
        True,
        True,
        False,
    ],
    detailled_status="PRESENT",
    assign=None,
    texts=[
        "Un lymphome de Hodgkin.",
        "Atteint d'un Waldenstörm",
        "Un LAGC",
        "anti LAGC: 10^4/mL",  # Dosage
    ],
)



import spacy
from pytest import mark
from spacy.pipeline.sentencizer import Sentencizer

from edsnlp.pipelines.core.sentences import SentenceSegmenter, terms

text = (
    "Le patient est admis pour des douleurs dans le bras droit. "
    "mais n'a pas de problème de locomotion. \n"
    "Historique d'AVC dans la famille\n"
    "Mais ne semble pas en être un\n"
    "Pourrait être un cas de rhume.\n"
    "Motif :\n"
    "Douleurs dans le bras droit."
)


@mark.parametrize("endlines", [True, False])
def test_sentences(nlp, endlines):
    nlp_blank = spacy.blank("fr")

    sentencizer = Sentencizer()
    segmenter = SentenceSegmenter(nlp, punct_chars=terms.punctuation, use_endlines=True)

    doc = nlp(text)
    doc_blank = nlp_blank(text)

    if endlines:
        doc_blank[28].tag_ = "EXCLUDED"

    doc_blank = segmenter(doc_blank)

    if endlines:
        assert len(list(doc_blank.sents)) == 6
    else:
        assert len(list(doc_blank.sents)) == 7
    assert len(list(sentencizer(doc).sents)) == 7

    segmenter(nlp_blank(""))


def test_false_positives(blank_nlp):
    false_positives = [
        "02.04.2018",
    ]

    for fp in false_positives:
        doc = blank_nlp(fp)
        assert len(list(doc.sents)) == 1



pseudo = [
    "aucun doute",
    "même si",
    "pas de condition",
    "pas de doute",
    "sans aucun doute",
    "sans condition",
    "sans risque",
]

confirmation = [
    "certain",
    "certaine",
    "certainement",
    "certaines",
    "certains",
    "confirmer",
    "évidemment",
    "évident",
    "évidente",
    "montrer que",
    "visiblement",
]

preceding = [
    "à condition",
    "à la condition que",
    "à moins que",
    "au cas où",
    "conditionnellement",
    "discret",
    "discrets",
    "discrète",
    "discrètes",
    "en admettant que",
    "en cas",
    "en considérant que",
    "en supposant que",
    "éventuel",
    "éventuels",
    "éventuelle",
    "éventuelles",
    "éventuellement",
    "exploration",
    "faudrait",
    "hypothèse",
    "hypothèses",
    "idée de",
    "pas confirmer",
    "pas sûr",
    "pas sûre",
    "peut correspondre",
    "peut-être",
    "peuvent correspondre",
    "possibilité",
    "possible",
    "possiblement",
    "potentiel",
    "potentielle",
    "potentiellement",
    "potentielles",
    "potentiels",
    "prédisposant à",
    "probable",
    "probablement",
    "probables",
    "recherche d'",
    "recherche de",
    "recherche des",
    "risque",
    "selon",
    "si",
    "s'il",
    "soit",
    "sous condition",
    "sous réserve",
    "suspicion",
]

following = [
    "?",
    "envisageable",
    "envisageables",
    "hypothétique",
    "hypothétiquement",
    "hypothétiques",
    "pas certain",
    "pas certaine",
    "pas clair",
    "pas claire",
    "pas confirmé",
    "pas confirmée",
    "pas confirmées",
    "pas confirmés",
    "pas évident",
    "pas évidente",
    "pas sûr",
    "pas sûre",
    "possibilité",
    "possible",
    "potentiel",
    "potentielle",
    "potentiels",
    "probable",
    "probables",
]

verbs_hyp = [
    "douter",
    "envisager",
    "explorer",
    "rechercher",
    "s'apparenter",
    "sembler",
    "soupçonner",
    "suggérer",
    "suspecter",
]

verbs_eds = [
    "abandonner",
    "abolir",
    "aborder",
    "accepter",
    "accidenter",
    "accompagnemer",
    "accompagner",
    "acoller",
    "acquérir",
    "activer",
    "actualiser",
    "adapter",
    "adhérer",
    "adjuver",
    "admettre",
    "administrer",
    "adopter",
    "adresser",
    "aggraver",
    "agir",
    "agréer",
    "aider",
    "aimer",
    "alcooliser",
    "alerter",
    "alimenter",
    "aller",
    "allonger",
    "alléger",
    "alterner",
    "altérer",
    "amender",
    "amener",
    "améliorer",
    "amyotrophier",
    "améliorer",
    "analyser",
    "anesthésier",
    "animer",
    "annexer",
    "annuler",
    "anonymiser",
    "anticiper",
    "anticoaguler",
    "apercevoir",
    "aplatir",
    "apparaître",
    "appareiller",
    "appeler",
    "appliquer",
    "apporter",
    "apprendre",
    "apprécier",
    "appuyer",
    "argumenter",
    "arquer",
    "arrêter",
    "arriver",
    "arrêter",
    "articuler",
    "aspirer",
    "asseoir",
    "assister",
    "associer",
    "assurer",
    "assécher",
    "attacher",
    "atteindre",
    "attendre",
    "attribuer",
    "augmenter",
    "autonomiser",
    "autoriser",
    "avaler",
    "avancer",
    "avertir",
    "avoir",
    "avérer",
    "aérer",
    "baisser",
    "ballonner",
    "blesser",
    "bloquer",
    "boire",
    "border",
    "brancher",
    "brûler",
    "bénéficier",
    "cadrer",
    "calcifier",
    "calculer",
    "calmer",
    "canaliser",
    "capter",
    "carencer",
    "casser",
    "centrer",
    "cerner",
    "certifier",
    "changer",
    "charger",
    "chevaucher",
    "choisir",
    "chronomoduler",
    "chuter",
    "cicatriser",
    "circoncire",
    "circuler",
    "classer",
    "codéiner",
    "coincer",
    "colorer",
    "combler",
    "commander",
    "commencer",
    "communiquer",
    "comparer",
    "compliquer",
    "compléter",
    "comporter",
    "comprendre",
    "comprimer",
    "concerner",
    "conclure",
    "condamner",
    "conditionner",
    "conduire",
    "confiner",
    "confirmer",
    "confronter",
    "congeler",
    "conjoindre",
    "conjuguer",
    "connaître",
    "connecter",
    "conseiller",
    "conserver",
    "considérer",
    "consommer",
    "constater",
    "constituer",
    "consulter",
    "contacter",
    "contaminer",
    "contenir",
    "contentionner",
    "continuer",
    "contracter",
    "contrarier",
    "contribuer",
    "contrôler",
    "convaincre",
    "convenir",
    "convier",
    "convoquer",
    "copier",
    "correspondre",
    "corriger",
    "corréler",
    "coucher",
    "coupler",
    "couvrir",
    "crapotter",
    "creuser",
    "croire",
    "croiser",
    "créer",
    "crémer",
    "crépiter",
    "cumuler",
    "curariser",
    "céder",
    "dater",
    "demander",
    "demeurer",
    "destiner",
    "devenir",
    "devoir",
    "diagnostiquer",
    "dialyser",
    "dicter",
    "diffuser",
    "différencier",
    "différer",
    "digérer",
    "dilater",
    "diluer",
    "diminuer",
    "diner",
    "dire",
    "diriger",
    "discuter",
    "disparaître",
    "disposer",
    "dissocier",
    "disséminer",
    "disséquer",
    "distendre",
    "distinguer",
    "divorcer",
    "documenter",
    "donner",
    "dorer",
    "doser",
    "doubler",
    "durer",
    "dyaliser",
    "dyspner",
    "débuter",
    "décaler",
    "déceler",
    "décider",
    "déclarer",
    "déclencher",
    "découvrir",
    "décrire",
    "décroître",
    "décurariser",
    "décéder",
    "dédier",
    "définir",
    "dégrader",
    "délivrer",
    "dépasser",
    "dépendre",
    "déplacer",
    "dépolir",
    "déposer",
    "dériver",
    "dérouler",
    "désappareiller",
    "désigner",
    "désinfecter",
    "désorienter",
    "détecter",
    "déterminer",
    "détruire",
    "développer",
    "dévouer",
    "dîner",
    "écraser",
    "effacer",
    "effectuer",
    "effondrer",
    "emboliser",
    "emmener",
    "empêcher",
    "encadrer",
    "encourager",
    "endormir",
    "endurer",
    "enlever",
    "enregistrer",
    "entamer",
    "entendre",
    "entourer",
    "entraîner",
    "entreprendre",
    "entrer",
    "envahir",
    "envisager",
    "envoyer",
    "espérer",
    "essayer",
    "estimer",
    "être",
    "examiner",
    "excentrer",
    "exciser",
    "exclure",
    "expirer",
    "expliquer",
    "explorer",
    "exposer",
    "exprimer",
    "extérioriser",
    "exécuter",
    "faciliter",
    "faire",
    "fatiguer",
    "favoriser",
    "faxer",
    "fermer",
    "figurer",
    "fixer",
    "focaliser",
    "foncer",
    "former",
    "fournir",
    "fractionner",
    "fragmenter",
    "fuiter",
    "fusionner",
    "garder",
    "graver",
    "guider",
    "gérer",
    "gêner",
    "honorer",
    "hopsitaliser",
    "hospitaliser",
    "hydrater",
    "hyperartérialiser",
    "hyperfixer",
    "hypertrophier",
    "hésiter",
    "identifier",
    "illustrer",
    "immuniser",
    "impacter",
    "implanter",
    "impliquer",
    "importer",
    "imposer",
    "impregner",
    "imprimer",
    "inclure",
    "indifferencier",
    "indiquer",
    "infecter",
    "infertiliser",
    "infirmer",
    "infiltrer",
    "informer",
    "inhaler",
    "initier",
    "injecter",
    "inscrire",
    "insister",
    "installer",
    "interdire",
    "interpréter",
    "interrompre",
    "intervenir",
    "intituler",
    "introduire",
    "intéragir",
    "inverser",
    "inviter",
    "ioder",
    "ioniser",
    "irradier",
    "itérativer",
    "joindre",
    "juger",
    "justifier",
    "laisser",
    "laminer",
    "lancer",
    "latéraliser",
    "laver",
    "lever",
    "lier",
    "ligaturer",
    "limiter",
    "lire",
    "localiser",
    "loger",
    "louper",
    "luire",
    "lutter",
    "lyricer",
    "lyser",
    "maculer",
    "macérer",
    "maintenir",
    "majorer",
    "malaiser",
    "manger",
    "manifester",
    "manipuler",
    "manquer",
    "marcher",
    "marier",
    "marmoner",
    "marquer",
    "masquer",
    "masser",
    "mater",
    "mener",
    "mesurer",
    "meteoriser",
    "mettre",
    "mitiger",
    "modifier",
    "moduler",
    "modérer",
    "monter",
    "montrer",
    "motiver",
    "moucheter",
    "mouler",
    "mourir",
    "multiopéréer",
    "munir",
    "muter",
    "médicaliser",
    "météoriser",
    "naître",
    "normaliser",
    "noter",
    "nuire",
    "numériser",
    "nécessiter",
    "négativer",
    "objectiver",
    "observer",
    "obstruer",
    "obtenir",
    "occasionner",
    "occuper",
    "opposer",
    "opérer",
    "organiser",
    "orienter",
    "ouvrir",
    "palper",
    "parasiter",
    "paraître",
    "parcourir",
    "parer",
    "paresthésier",
    "parfaire",
    "partager",
    "partir",
    "parvenir",
    "passer",
    "penser",
    "percevoir",
    "perdre",
    "perforer",
    "permettre",
    "persister",
    "personnaliser",
    "peser",
    "pigmenter",
    "piloter",
    "placer",
    "plaindre",
    "planifier",
    "plier",
    "plonger",
    "porter",
    "poser",
    "positionner",
    "posséder",
    "poursuivre",
    "pousser",
    "pouvoir",
    "pratiquer",
    "preciser",
    "prendre",
    "prescrire",
    "prier",
    "produire",
    "programmer",
    "prolonger",
    "prononcer",
    "proposer",
    "prouver",
    "provoquer",
    "préciser",
    "précéder",
    "prédominer",
    "préexister",
    "préférer",
    "prélever",
    "préparer",
    "présenter",
    "préserver",
    "prévenir",
    "prévoir",
    "puruler",
    "pénétrer",
    "radiofréquencer",
    "ralentir",
    "ramener",
    "rappeler",
    "rapporter",
    "rapprocher",
    "rassurer",
    "rattacher",
    "rattraper",
    "realiser",
    "recenser",
    "recevoir",
    "rechercher",
    "recommander",
    "reconnaître",
    "reconsulter",
    "recontacter",
    "recontrôler",
    "reconvoquer",
    "recouvrir",
    "recueillir",
    "recuperer",
    "redescendre",
    "rediscuter",
    "refaire",
    "refouler",
    "refuser",
    "regarder",
    "rehausser",
    "relancer",
    "relayer",
    "relever",
    "relire",
    "relâcher",
    "remanier",
    "remarquer",
    "remercier",
    "remettre",
    "remonter",
    "remplacer",
    "remplir",
    "rencontrer",
    "rendormir",
    "rendre",
    "renfermer",
    "renforcer",
    "renouveler",
    "renseigner",
    "rentrer",
    "reparler",
    "repasser",
    "reporter",
    "reprendre",
    "represcrire",
    "reproduire",
    "reprogrammer",
    "représenter",
    "repérer",
    "requérir",
    "respecter",
    "ressembler",
    "ressentir",
    "rester",
    "restreindre",
    "retarder",
    "retenir",
    "retirer",
    "retrouver",
    "revasculariser",
    "revenir",
    "reverticaliser",
    "revoir",
    "rompre",
    "rouler",
    "réadapter",
    "réadmettre",
    "réadresser",
    "réaliser",
    "récidiver",
    "récupérer",
    "rédiger",
    "réduire",
    "réessayer",
    "réexpliquer",
    "référer",
    "régler",
    "régresser",
    "réhausser",
    "réopérer",
    "répartir",
    "répondre",
    "répéter",
    "réserver",
    "résorber",
    "résoudre",
    "réséquer",
    "réveiller",
    "révéler",
    "réévaluer",
    "rêver",
    "sacrer",
    "saisir",
    "satisfaire",
    "savoir",
    "scanner",
    "scolariser",
    "sembler",
    "sensibiliser",
    "sentir",
    "serrer",
    "servir",
    "sevrer",
    "signaler",
    "signer",
    "situer",
    "siéger",
    "soigner",
    "sommeiller",
    "sonder",
    "sortir",
    "souffler",
    "souhaiter",
    "soulager",
    "soussigner",
    "souvenir",
    "spécialiser",
    "stabiliser",
    "statuer",
    "stenter",
    "stopper",
    "stratifier",
    "subir",
    "substituer",
    "sucrer",
    "suggérer",
    "suivre",
    "supporter",
    "supprimer",
    "surajouter",
    "surmonter",
    "surveiller",
    "survenir",
    "suspecter",
    "suspendre",
    "suturer",
    "synchroniser",
    "systématiser",
    "sécréter",
    "sécuriser",
    "sédater",
    "séjourner",
    "séparer",
    "taire",
    "taper",
    "teinter",
    "tendre",
    "tenir",
    "tenter",
    "terminer",
    "tester",
    "thromboser",
    "tirer",
    "tiroir",
    "tissulaire",
    "titulariser",
    "tolérer",
    "tourner",
    "tracer",
    "trachéotomiser",
    "traduire",
    "traiter",
    "transcrire",
    "transférer",
    "transmettre",
    "transporter",
    "trasnfixer",
    "travailler",
    "tronquer",
    "trouver",
    "téléphoner",
    "ulcérer",
    "uriner",
    "utiliser",
    "vacciner",
    "valider",
    "valoir",
    "varier",
    "vasculariser",
    "venir",
    "verifier",
    "vieillir",
    "viser",
    "visualiser",
    "vivre",
    "voir",
    "vouloir",
    "vérifier",
    "ébaucher",
    "écarter",
    "échographier",
    "échoguider",
    "échoir",
    "échouer",
    "éclairer",
    "écraser",
    "élargir",
    "éliminer",
    "émousser",
    "épaissir",
    "épargner",
    "épuiser",
    "épurer",
    "équilibrer",
    "établir",
    "étager",
    "étendre",
    "étiqueter",
    "étrangler",
    "évaluer",
    "éviter",
    "évoluer",
    "évoquer",
    "être",
]



import pytest

pytest.importorskip("umls_downloader", reason="umls_downloader package not found")

import os  # noqa: E402
from pathlib import Path  # noqa: E402

from edsnlp.core import PipelineProtocol  # noqa: E402
from edsnlp.pipelines.ner.umls.patterns import get_path, get_patterns  # noqa: E402
from edsnlp.utils.examples import parse_example  # noqa: E402

examples = [
    "Grosse <ent umls=C0010200>toux</ent> : Le malade a été mordu "
    "<ent umls=C1257901>par</ent> des <ent umls=C0002668>Amphibiens</ent>"
    " sous le <ent umls=C0022742>genou</ent>"
]
pattern_config = {"lang": ["FRE"], "sources": ["MSHFRE"]}


@pytest.mark.skipif(not os.getenv("UMLS_API_KEY"), reason="No UMLS_API_KEY given")
def test_get_patterns():
    path, _, _ = get_path(pattern_config)
    assert isinstance(path, Path)

    api_key = os.getenv("UMLS_API_KEY")

    try:
        patterns = get_patterns(pattern_config)
    except AttributeError:
        if not api_key:
            raise Exception(
                "Missing UMLS_API_KEY env variable to download the UMLS. "
                f"UMLS cannot be found at path: {path}.\nTo skip the test, "
                f"uninstall the umls_downloader package."
            )
        raise

    assert path.exists()

    assert len(patterns) == 48587


@pytest.mark.skipif(not os.getenv("UMLS_API_KEY"), reason="No UMLS_API_KEY given")
def test_add_pipe(blank_nlp: PipelineProtocol):
    path, _, _ = get_path(pattern_config)
    if not path.exists():
        pytest.xfail(
            "The umls_downloader package is installed but "
            f"UMLS cannot be found at path: {path}"
        )

    blank_nlp.add_pipe("eds.umls", config=dict(pattern_config=pattern_config))

    for text, entities in map(parse_example, examples):
        doc = blank_nlp(text)
        assert len(doc.ents) == len(entities)

        for ent, entity in zip(doc.ents, entities):
            assert ent.text == text[entity.start_char : entity.end_char]
            assert ent._.umls == ent.kb_id_ == entity.modifiers[0].value



from ..terms import ASYMPTOMATIC

main_pattern = dict(
    source="main",
    regex=[
        r"defaillance.{1,10}cardi",
        r"(œ|oe)deme.{1,10}pulmon",
        r"(œ|oe)deme.{1,10}poumon",
        r"decompensation.{1,10}card",
        r"choc.{1,30}cardio",
        r"greffe.{1,10}c(œ|oe)ur",
        r"greffe.{1,10}cardia",
        r"transplantation.{1,10}c(œ|oe)ur",
        r"transplantation.{1,10}cardia",
        r"arret.{1,10}cardi",
        r"c(œ|oe)ur pulmo",
        r"foie.card",
        r"pace.?maker",
        r"stimulateur.cardiaque",
        r"valve.{1,30}(meca|artific)",
    ],
    regex_attr="NORM",
)

symptomatic = dict(
    source="symptomatic",
    regex=[
        r"cardiopathi",
        r"cardiomyopathi",
        r"d(i|y)sfonction.{1,15}(ventricul|\bvg|cardiaque)",
        r"valvulopathie",
        r"\bic\b.{1,10}(droite|gauche)",
    ],
    regex_attr="NORM",
    exclude=dict(
        regex=ASYMPTOMATIC + [r"(?<!\bnon.)ischem"],  # Exclusion of ischemic events
        window=5,
    ),
)

with_minimum_severity = dict(
    source="min_severity",
    regex=[
        r"insuffisance.{1,10}(\bcardi|\bdiasto|\bventri|\bmitral|tri.?cusp)",
        r"(retrecissement|stenose).(aortique|mitral)",
        r"\brac\b",
        r"\brm\b",
    ],
    regex_attr="NORM",
    exclude=dict(
        regex=ASYMPTOMATIC + ["minime", "modere", r"non.serre"],
        window=5,
    ),
)

acronym = dict(
    source="acronym",
    regex=[
        r"\bOAP\b",
        r"\bCMH\b",
    ],
    regex_attr="TEXT",
)

AF_main_pattern = dict(
    source="AF_main",
    regex=[
        r"fibrill?ation.{1,3}(atriale|auriculaire|ventriculaire)",
        r"flutter",
        r"brady.?arythmie",
        r"pace.?maker",
    ],
)

AF_acronym = dict(
    source="AF_acronym",
    regex=[
        r"\bFA\b",
        r"\bAC.?FA\b",
    ],
    regex_attr="TEXT",
)

default_patterns = [
    main_pattern,
    symptomatic,
    acronym,
    AF_main_pattern,
    AF_acronym,
    with_minimum_severity,
]



results_copd = dict(
    has_match=[
        True,
        True,
        True,
        False,
        False,
        True,
    ],
    detailled_status="PRESENT",
    assign=None,
    texts=[
        "Une fibrose interstitielle diffuse idiopathique",
        "Patient atteint de pneumoconiose",
        "Présence d'une HTAP.",
        "On voit une hypertension pulmonaire minime",
        "La patiente a été mis sous oxygénorequérance",  # Ponctual: not extracted
        "La patiente est sous oxygénorequérance au long cours",
    ],
)
